from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterable


def identity(arg: Any | None = None, *args, **kwargs) -> Any:
    ...


def inject_method(self,
                  func: Callable[..., Any],
                  name: str | None = None) -> None:
    ...


def compatible(config: Dict[str, Any],
               func: Callable,
               start: int = 0,
               keywords: bool | Iterable[str] = True) -> Dict[str, Any]:
    ...



from os import PathLike
from typing import Iterable
from typing import List
from collections.abc import Generator

WIN32: bool
LINUX: bool
FREEBSD: bool
DARWIN: bool
POSIX: bool


def platform_data_dir() -> str:
    ...


def platform_config_dir() -> str:
    ...


def platform_cache_dir() -> str:
    ...


def get_app_data_dir(appname: str, *args) -> str:
    ...


def ensure_app_data_dir(appname: str, *args) -> str:
    ...


def get_app_config_dir(appname: str, *args) -> str:
    ...


def ensure_app_config_dir(appname: str, *args) -> str:
    ...


def get_app_cache_dir(appname: str, *args) -> str:
    ...


def ensure_app_cache_dir(appname: str, *args) -> str:
    ...


def find_exe(
    name: str | PathLike,
    multi: bool = False,
    path: str | PathLike | Iterable[str | PathLike] | None = None
) -> str | List[str] | None:
    ...


def find_path(name: str | PathLike,
              path: str | Iterable[str | PathLike] | None = None,
              exact: bool = False) -> Generator[str, None, None]:
    ...



from typing import Tuple
from ubelt.util_const import NoParamType
from typing import List
from typing import TypeVar

T = TypeVar("T")


def argval(key: str | Tuple[str, ...],
           default: T | NoParamType = ...,
           argv: List[str] | None = None) -> str | T:
    ...


def argflag(key: str | Tuple[str, ...], argv: List[str] | None = None) -> bool:
    ...



"""
Utility functions for manipulating iterables, lists, and sequences.

The :func:`chunks` function splits a list into smaller parts. There are different strategies for how to do this.

The :func:`flatten` function take a list of lists and removes the inner lists. This
only removes one level of nesting.

The :func:`iterable` function checks if an object is iterable or not. Similar to the
:func:`callable` builtin function.

The :func:`argmax`, :func:`argmin`, and :func:`argsort` work similarly to the
analogous :mod:`numpy` functions, except they operate on dictionaries and other
Python builtin types.

The :func:`take` and :func:`compress` are generators, and also similar to their
lesser known, but very useful numpy equivalents.

There are also other numpy inspired functions: :func:`unique`,
:func:`argunique`, :func:`unique_flags`, and :func:`boolmask`.
"""
import itertools as it
import math
import operator
from collections import abc as collections_abc
from itertools import zip_longest
from ubelt import util_const
from ubelt import util_dict

__all__ = [
    'allsame', 'argmax', 'argmin', 'argsort', 'argunique', 'boolmask',
    'chunks', 'compress', 'flatten', 'iter_window', 'iterable', 'peek', 'take',
    'unique', 'unique_flags',
]


class chunks(object):
    """
    Generates successive n-sized chunks from ``items``.

    If the last chunk has less than n elements, ``bordermode`` is used to
    determine fill values.

    Note:
        FIXME:
            When nchunks is given, that's how many chunks we should get
            but the issue is that chunksize is not well defined in that instance
            For instance how do we turn a list with 4 elements into 3 chunks
            where does the extra item go?

        In ubelt <= 0.10.3 there is a bug when specifying nchunks,
        where it chooses a chunksize that is too large. Specify
        ``legacy=True`` to get the old buggy behavior if needed.

    Notes:
        This is similar to functionality provided by
            :func:`more_itertools.chunked`,
            :func:`more_itertools.chunked_even`,
            :func:`more_itertools.sliced`,
            :func:`more_itertools.divide`,

    Yields:
        List[T]:
            subsequent non-overlapping chunks of the input items

    Attributes:
        remainder (int): number of leftover items that don't divide cleanly

    References:
        .. [SO_434287] http://stackoverflow.com/questions/434287/iterate-over-a-list-in-chunks

    Example:
        >>> import ubelt as ub
        >>> items = '1234567'
        >>> genresult = ub.chunks(items, chunksize=3)
        >>> list(genresult)
        [['1', '2', '3'], ['4', '5', '6'], ['7']]

    Example:
        >>> import ubelt as ub
        >>> items = [1, 2, 3, 4, 5, 6, 7]
        >>> genresult = ub.chunks(items, chunksize=3, bordermode='none')
        >>> assert list(genresult) == [[1, 2, 3], [4, 5, 6], [7]]
        >>> genresult = ub.chunks(items, chunksize=3, bordermode='cycle')
        >>> assert list(genresult) == [[1, 2, 3], [4, 5, 6], [7, 1, 2]]
        >>> genresult = ub.chunks(items, chunksize=3, bordermode='replicate')
        >>> assert list(genresult) == [[1, 2, 3], [4, 5, 6], [7, 7, 7]]

    Example:
        >>> import ubelt as ub
        >>> assert len(list(ub.chunks(range(2), nchunks=2))) == 2
        >>> assert len(list(ub.chunks(range(3), nchunks=2))) == 2
        >>> # Note: ub.chunks will not do the 2,1,1 split
        >>> assert len(list(ub.chunks(range(4), nchunks=3))) == 3
        >>> assert len(list(ub.chunks([], 2, bordermode='none'))) == 0
        >>> assert len(list(ub.chunks([], 2, bordermode='cycle'))) == 0
        >>> assert len(list(ub.chunks([], 2, None, bordermode='replicate'))) == 0

    Example:
        >>> from ubelt.util_list import *  # NOQA
        >>> def _check_len(self):
        ...     assert len(self) == len(list(self))
        >>> _check_len(chunks(list(range(3)), nchunks=2))
        >>> _check_len(chunks(list(range(2)), nchunks=2))
        >>> _check_len(chunks(list(range(2)), nchunks=3))

    Example:
        >>> from ubelt.util_list import *  # NOQA
        >>> import pytest
        >>> assert pytest.raises(ValueError, chunks, range(9))
        >>> assert pytest.raises(ValueError, chunks, range(9), chunksize=2, nchunks=2)
        >>> assert pytest.raises(TypeError, len, chunks((_ for _ in range(2)), 2))

    Example:
        >>> from ubelt.util_list import *  # NOQA
        >>> import ubelt as ub
        >>> basis = {
        >>>     'legacy': [False, True],
        >>>     'chunker': [{'nchunks': 3}, {'nchunks': 4}, {'nchunks': 5}, {'nchunks': 7}, {'chunksize': 3}],
        >>>     'items': [range(2), range(4), range(5), range(7), range(9)],
        >>>     'bordermode': ['none', 'cycle', 'replicate'],
        >>> }
        >>> grid_items = list(ub.named_product(basis))
        >>> rows = []
        >>> for grid_item in ub.ProgIter(grid_items):
        >>>     chunker = grid_item.get('chunker')
        >>>     grid_item.update(chunker)
        >>>     kw = ub.dict_diff(grid_item, {'chunker'})
        >>>     self = chunk_iter = ub.chunks(**kw)
        >>>     chunked = list(chunk_iter)
        >>>     chunk_lens = list(map(len, chunked))
        >>>     row = ub.dict_union(grid_item, {'chunk_lens': chunk_lens, 'chunks': chunked})
        >>>     row['chunker'] = str(row['chunker'])
        >>>     if not row['legacy'] and 'nchunks' in kw:
        >>>         assert kw['nchunks'] == row['nchunks']
        >>>     row.update(chunk_iter.__dict__)
        >>>     rows.append(row)
        >>> # xdoctest: +SKIP
        >>> import pandas as pd
        >>> df = pd.DataFrame(rows)
        >>> for _, subdf in df.groupby('chunker'):
        >>>     print(subdf)

    """
    def __init__(self, items, chunksize=None, nchunks=None, total=None,
                 bordermode='none', legacy=False):
        """
        Args:
            items (Iterable): input to iterate over

            chunksize (int | None): size of each sublist yielded

            nchunks (int | None): number of chunks to create (
                cannot be specified if chunksize is specified)

            bordermode (str): determines how to handle the last case if the
                length of the input is not divisible by chunksize valid values
                are: {'none', 'cycle', 'replicate'}

            total (int | None): hints about the length of the input

            legacy (bool):
                if True use old behavior, defaults to False. This will be
                removed in the future.
        """
        if nchunks is not None and chunksize is not None:  # nocover
            raise ValueError('Cannot specify both chunksize and nchunks')
        if nchunks is None and chunksize is None:  # nocover
            raise ValueError('Must specify either chunksize or nchunks')

        if total is None:
            try:
                total = len(items)
            except TypeError:
                pass  # iterators dont know len

        if bordermode is None:  # nocover
            bordermode = 'none'

        if nchunks is None:
            if total is not None:
                nchunks = int(math.ceil(total / chunksize))
            remainder = 0
        else:
            if total is None:
                raise ValueError(
                    'Need to specify total to use nchunks on an iterable '
                    'without length hints')
            if legacy:
                chunksize: int = int(math.ceil(total / nchunks))
                remainder = 0
            else:
                if bordermode == 'none':
                    # I feel like this could be simpler
                    chunksize: int = max(int(math.floor(total / nchunks)), 1)
                    nchunks: int = min(int(math.ceil(total / chunksize)), nchunks)
                    chunked_total: int = chunksize * nchunks
                    remainder: int = total - chunked_total
                else:
                    # not working
                    chunksize: int = max(int(math.ceil(total / nchunks)), 1)
                    # Can artificially extend the size in this case
                    # total = chunksize * nchunks
                    remainder = 0

        self.legacy = legacy
        self.remainder: int = remainder
        self.items = items
        self.total = total
        self.nchunks = nchunks
        self.chunksize = chunksize
        self.bordermode = bordermode

    def __len__(self):
        if self.nchunks is None:
            raise TypeError('length is unknown')
        return self.nchunks

    def __iter__(self):
        bordermode = self.bordermode
        items = self.items
        chunksize = self.chunksize

        if not self.legacy and self.nchunks is not None:
            return self._new_iterator()
        else:
            if bordermode is None or bordermode == 'none':
                return self.noborder(items, chunksize)
            elif bordermode == 'cycle':
                return self.cycle(items, chunksize)
            elif bordermode == 'replicate':
                return self.replicate(items, chunksize)
            else:
                raise ValueError('unknown bordermode=%r' % (bordermode,))

    def _new_iterator(self):
        chunksize = self.chunksize
        nchunks = self.nchunks
        chunksize = self.chunksize
        remainder = self.remainder

        if self.bordermode == 'cycle':
            iterator = it.cycle(iter(self.items))
        elif self.bordermode == 'replicate':
            def replicator(items):
                for item in items:
                    yield item
                while True:
                    yield item
            iterator = replicator(iter(self.items))
        elif self.bordermode == 'none':
            iterator = iter(self.items)
        else:
            raise KeyError(self.bordermode)

        # Build an iterator that describes how big each chunk will be
        if remainder:
            # TODO:
            # handle replicate and cycle border modes
            # TODO:
            # benchmark different methods
            chunksize_iter = it.chain(
                it.repeat(chunksize + 1, remainder),
                it.repeat(chunksize, nchunks - remainder)
            )
        else:
            chunksize_iter = it.repeat(chunksize, nchunks)
        for _chunksize in chunksize_iter:
            chunk = list(it.islice(iterator, _chunksize))
            # if chunk:
            yield chunk

    @staticmethod
    def noborder(items, chunksize):
        # feed the same iter to zip_longest multiple times, this causes it to
        # consume successive values of the same sequence
        sentinel = object()
        copied_iters = [iter(items)] * chunksize
        chunks_with_sentinals = zip_longest(*copied_iters, fillvalue=sentinel)
        # Dont fill empty space in the last chunk, just return it as is
        for chunk in chunks_with_sentinals:
            yield [item for item in chunk if item is not sentinel]

    @staticmethod
    def cycle(items, chunksize):
        sentinel = object()
        copied_iters = [iter(items)] * chunksize
        chunks_with_sentinals = zip_longest(*copied_iters, fillvalue=sentinel)
        # Fill empty space in the last chunk with values from the beginning
        bordervalues = it.cycle(iter(items))
        for chunk in chunks_with_sentinals:
            yield [item if item is not sentinel else next(bordervalues)
                   for item in chunk]

    @staticmethod
    def replicate(items, chunksize):
        sentinel = object()
        copied_iters = [iter(items)] * chunksize
        # Fill empty space in the last chunk by replicating the last value
        chunks_with_sentinals = zip_longest(*copied_iters, fillvalue=sentinel)
        for chunk in chunks_with_sentinals:
            filt_chunk = [item for item in chunk if item is not sentinel]
            if len(filt_chunk) == chunksize:
                yield filt_chunk
            else:
                sizediff = (chunksize - len(filt_chunk))
                padded_chunk = filt_chunk + [filt_chunk[-1]] * sizediff
                yield padded_chunk


def iterable(obj, strok=False):
    """
    Checks if the input implements the iterator interface. An exception is made
    for strings, which return False unless ``strok`` is True

    Args:
        obj (object): a scalar or iterable input

        strok (bool):
            if True allow strings to be interpreted as iterable.
            Defaults to False.

    Returns:
        bool: True if the input is iterable

    Example:
        >>> import ubelt as ub
        >>> obj_list = [3, [3], '3', (3,), [3, 4, 5], {}]
        >>> result = [ub.iterable(obj) for obj in obj_list]
        >>> assert result == [False, True, False, True, True, True]
        >>> result = [ub.iterable(obj, strok=True) for obj in obj_list]
        >>> assert result == [False, True, True, True, True, True]
    """
    try:
        iter(obj)
    except Exception:
        return False
    else:
        return strok or not isinstance(obj, str)


def take(items, indices, default=util_const.NoParam):
    """
    Lookup a subset of an indexable object using a sequence of indices.

    The ``items`` input is usually a list or dictionary. When ``items`` is a
    list, this should be a sequence of integers. When ``items`` is a dict, this
    is a list of keys to lookup in that dictionary.

    For dictionaries, a default may be specified as a placeholder to use if a
    key from ``indices`` is not in ``items``.

    Args:
        items (Sequence[VT] | Mapping[KT, VT]):
            An indexable object to select items from.

        indices (Iterable[int | KT]):
            A sequence of indexes into ``items``.

        default (Any | NoParamType):
            if specified ``items`` must support the ``get`` method and
            this will be used as the default value.

    Yields:
        VT: a selected item within the list

    SeeAlso:
        :func:`ubelt.dict_subset`

    Note:
        ``ub.take(items, indices)`` is equivalent to
        ``(items[i] for i in indices)`` when ``default`` is unspecified.

    Notes:
        This is based on the :func:`numpy.take` function, but written in pure
        python.

        Do not confuse this with :func:`more_itertools.take`, the behavior is
        very different.

    Example:
        >>> import ubelt as ub
        >>> items = [0, 1, 2, 3]
        >>> indices = [2, 0]
        >>> list(ub.take(items, indices))
        [2, 0]

    Example:
        >>> import ubelt as ub
        >>> dict_ = {1: 'a', 2: 'b', 3: 'c'}
        >>> keys = [1, 2, 3, 4, 5]
        >>> result = list(ub.take(dict_, keys, None))
        >>> assert result == ['a', 'b', 'c', None, None]

    Example:
        >>> import ubelt as ub
        >>> dict_ = {1: 'a', 2: 'b', 3: 'c'}
        >>> keys = [1, 2, 3, 4, 5]
        >>> try:
        >>>     print(list(ub.take(dict_, keys)))
        >>>     raise AssertionError('did not get key error')
        >>> except KeyError:
        >>>     print('correctly got key error')
    """
    if default is util_const.NoParam:
        for index in indices:
            yield items[index]
    else:
        for index in indices:
            yield items.get(index, default)


def compress(items, flags):
    """
    Selects from ``items`` where the corresponding value in ``flags`` is True.


    Args:
        items (Iterable[Any]): a sequence to select items from

        flags (Iterable[bool]): corresponding sequence of bools

    Returns:
        Iterable[Any]: a subset of masked items

    Notes:
        This function is based on :func:`numpy.compress`, but is pure Python
        and swaps the condition and array argument to be consistent with
        :func:`ubelt.take`.

        This is equivalent to :func:`itertools.compress`.

    Example:
        >>> import ubelt as ub
        >>> items = [1, 2, 3, 4, 5]
        >>> flags = [False, True, True, False, True]
        >>> list(ub.compress(items, flags))
        [2, 3, 5]
    """
    return it.compress(items, flags)


def flatten(nested):
    """
    Transforms a nested iterable into a flat iterable.

    Args:
        nested (Iterable[Iterable[Any]]): list of lists

    Returns:
        Iterable[Any]: flattened items

    Notes:
        Equivalent to :func:`more_itertools.flatten` and
        :func:`itertools.chain.from_iterable`.

    Example:
        >>> import ubelt as ub
        >>> nested = [['a', 'b'], ['c', 'd']]
        >>> list(ub.flatten(nested))
        ['a', 'b', 'c', 'd']
    """
    return it.chain.from_iterable(nested)


def unique(items, key=None):
    """
    Generates unique items in the order they appear.

    Args:
        items (Iterable[T]): list of items

        key (Callable[[T], Any] | None):
            Custom normalization function.
            If specified, this function generates items where ``key(item)`` is
            unique.

    Yields:
        T:
            a unique item from the input sequence

    Notes:
        Functionally equivalent to :func:`more_itertools.unique_everseen`.

    Example:
        >>> import ubelt as ub
        >>> items = [4, 6, 6, 0, 6, 1, 0, 2, 2, 1]
        >>> unique_items = list(ub.unique(items))
        >>> assert unique_items == [4, 6, 0, 1, 2]

    Example:
        >>> import ubelt as ub
        >>> items = ['A', 'a', 'b', 'B', 'C', 'c', 'D', 'e', 'D', 'E']
        >>> unique_items = list(ub.unique(items, key=str.lower))
        >>> assert unique_items == ['A', 'b', 'C', 'D', 'e']
        >>> unique_items = list(ub.unique(items))
        >>> assert unique_items == ['A', 'a', 'b', 'B', 'C', 'c', 'D', 'e', 'E']
    """
    seen = set()
    if key is None:
        for item in items:
            if item not in seen:
                seen.add(item)
                yield item
    else:
        for item in items:
            norm = key(item)
            if norm not in seen:
                seen.add(norm)
                yield item


def argunique(items, key=None):
    """
    Returns indices corresponding to the first instance of each unique item.

    Args:
        items (Sequence[VT]): indexable collection of items

        key (Callable[[VT], Any] | None):
            Custom normalization function.
            If specified, this function generates indexes where
            ``key(item[index])`` is unique.

    Returns:
        Iterator[int] : indices of the unique items

    Example:
        >>> import ubelt as ub
        >>> items = [0, 2, 5, 1, 1, 0, 2, 4]
        >>> indices = list(ub.argunique(items))
        >>> assert indices == [0, 1, 2, 3, 7]
        >>> indices = list(ub.argunique(items, key=lambda x: x % 2 == 0))
        >>> assert indices == [0, 2]
    """
    if key is None:
        return unique(range(len(items)), key=lambda i: items[i])
    else:
        return unique(range(len(items)), key=lambda i: key(items[i]))


def unique_flags(items, key=None):
    """
    Returns a list of booleans corresponding to the first instance of each
    unique item.

    Args:
        items (Sequence[VT]): indexable collection of items

        key (Callable[[VT], Any] | None):
            Custom normalization function.
            If specified generates True if ``key(item)`` is unique and False
            otherwise.

    Returns:
        List[bool] : flags the items that are unique

    Example:
        >>> import ubelt as ub
        >>> items = [0, 2, 1, 1, 0, 9, 2]
        >>> flags = ub.unique_flags(items)
        >>> assert flags == [True, True, True, False, False, True, False]
        >>> flags = ub.unique_flags(items, key=lambda x: x % 2 == 0)
        >>> assert flags == [True, False, True, False, False, False, False]
    """
    len_ = len(items)
    if key is None:
        item_to_index = dict(zip(reversed(items), reversed(range(len_))))
        indices = item_to_index.values()
    else:
        indices = argunique(items, key=key)
    flags = boolmask(indices, len_)
    return flags


def boolmask(indices, maxval=None):
    """
    Constructs a list of booleans where an item is True if its position is in
    ``indices`` otherwise it is False.

    Args:
        indices (List[int]): list of integer indices

        maxval (int | None):
            length of the returned list. If not specified this is inferred
            using ``max(indices)``

    Returns:
        List[bool]:
            mask - a list of booleans. mask[idx] is True if idx in indices

    Note:
        In the future the arg ``maxval`` may change its name to ``shape``

    Example:
        >>> import ubelt as ub
        >>> indices = [0, 1, 4]
        >>> mask = ub.boolmask(indices, maxval=6)
        >>> assert mask == [True, True, False, False, True, False]
        >>> mask = ub.boolmask(indices)
        >>> assert mask == [True, True, False, False, True]
    """
    if maxval is None:
        indices = list(indices)
        maxval = max(indices) + 1
    mask = [False] * maxval
    for index in indices:
        mask[index] = True
    return mask


def iter_window(iterable, size=2, step=1, wrap=False):
    """
    Iterates through iterable with a window size. This is essentially a 1D
    sliding window.

    Args:
        iterable (Iterable[T]): an iterable sequence

        size (int): Sliding window size. Defaults to 2.

        step (int): Sliding step size. Default to 1.

        wrap (bool):
            If True, the last window will "wrap-around" to include items from
            the start of the input sequence in order to always produce
            consistently sized chunks. Otherwise, the last chunk may be smaller
            if there are not enough items in the sequence.. Defaults to False.

    Returns:
        Iterable[T]: returns a possibly overlapping windows in a sequence

    Notes:
        Similar to :func:`more_itertools.windowed`,
        Similar to :func:`more_itertools.pairwise`,
        Similar to :func:`more_itertools.triplewise`,
        Similar to :func:`more_itertools.sliding_window`

    Example:
        >>> import ubelt as ub
        >>> iterable = [1, 2, 3, 4, 5, 6]
        >>> size, step, wrap = 3, 1, True
        >>> window_iter = ub.iter_window(iterable, size, step, wrap)
        >>> window_list = list(window_iter)
        >>> print('window_list = %r' % (window_list,))
        window_list = [(1, 2, 3), (2, 3, 4), (3, 4, 5), (4, 5, 6), (5, 6, 1), (6, 1, 2)]

    Example:
        >>> import ubelt as ub
        >>> iterable = [1, 2, 3, 4, 5, 6]
        >>> size, step, wrap = 3, 2, True
        >>> window_iter = ub.iter_window(iterable, size, step, wrap)
        >>> window_list = list(window_iter)
        >>> print('window_list = {!r}'.format(window_list))
        window_list = [(1, 2, 3), (3, 4, 5), (5, 6, 1)]

    Example:
        >>> import ubelt as ub
        >>> iterable = [1, 2, 3, 4, 5, 6]
        >>> size, step, wrap = 3, 2, False
        >>> window_iter = ub.iter_window(iterable, size, step, wrap)
        >>> window_list = list(window_iter)
        >>> print('window_list = {!r}'.format(window_list))
        window_list = [(1, 2, 3), (3, 4, 5)]

    Example:
        >>> import ubelt as ub
        >>> iterable = []
        >>> size, step, wrap = 3, 2, False
        >>> window_iter = ub.iter_window(iterable, size, step, wrap)
        >>> window_list = list(window_iter)
        >>> print('window_list = {!r}'.format(window_list))
        window_list = []
    """
    # it.tee may be slow, but works on all iterables
    iter_list = it.tee(iterable, size)
    if wrap:
        # Secondary iterables need to be cycled for wraparound
        iter_list = [iter_list[0]] + list(map(it.cycle, iter_list[1:]))
    # Step each iterator the appropriate number of times
    try:
        for count, iter_ in enumerate(iter_list[1:], start=1):
            for _ in range(count):
                next(iter_)
    except StopIteration:
        return iter(())
    else:
        _window_iter = zip(*iter_list)
        # Account for the step size
        window_iter = it.islice(_window_iter, 0, None, step)
        return window_iter


def allsame(iterable, eq=operator.eq):
    """
    Determine if all items in a sequence are the same

    Args:
        iterable (Iterable[T]):
            items to determine if they are all the same

        eq (Callable[[T, T], bool]):
            function used to test for equality.
            Defaults to :func:`operator.eq`.

    Returns:
        bool: True if all items are equal, otherwise False

    Notes:
        Similar to :func:`more_itertools.all_equal`

    Example:
        >>> import ubelt as ub
        >>> ub.allsame([1, 1, 1, 1])
        True
        >>> ub.allsame([])
        True
        >>> ub.allsame([0, 1])
        False
        >>> iterable = iter([0, 1, 1, 1])
        >>> next(iterable)
        >>> ub.allsame(iterable)
        True
        >>> ub.allsame(range(10))
        False
        >>> ub.allsame(range(10), lambda a, b: True)
        True
    """
    iter_ = iter(iterable)
    try:
        first = next(iter_)
    except StopIteration:
        return True
    return all(eq(first, item) for item in iter_)


def argsort(indexable, key=None, reverse=False):
    """
    Returns the indices that would sort a indexable object.

    This is similar to :func:`numpy.argsort`, but it is written in pure python
    and works on both lists and dictionaries.

    Args:
        indexable (Iterable[VT] | Mapping[KT, VT]): indexable to sort by

        key (Callable[[VT], VT] | None):
            If specified, customizes the ordering of the indexable.

        reverse (bool): if True returns in descending order. Default to False.

    Returns:
        List[int] | List[KT]:
            indices - list of indices that sorts the indexable

    Example:
        >>> import ubelt as ub
        >>> # argsort works on dicts by returning keys
        >>> dict_ = {'a': 3, 'b': 2, 'c': 100}
        >>> indices = ub.argsort(dict_)
        >>> assert list(ub.take(dict_, indices)) == sorted(dict_.values())
        >>> # argsort works on lists by returning indices
        >>> indexable = [100, 2, 432, 10]
        >>> indices = ub.argsort(indexable)
        >>> assert list(ub.take(indexable, indices)) == sorted(indexable)
        >>> # Can use iterators, but be careful. It exhausts them.
        >>> indexable = reversed(range(100))
        >>> indices = ub.argsort(indexable)
        >>> assert indices[0] == 99
        >>> # Can use key just like sorted
        >>> indexable = [[0, 1, 2], [3, 4], [5]]
        >>> indices = ub.argsort(indexable, key=len)
        >>> assert indices == [2, 1, 0]
        >>> # Can use reverse just like sorted
        >>> indexable = [0, 2, 1]
        >>> indices = ub.argsort(indexable, reverse=True)
        >>> assert indices == [1, 2, 0]
    """
    # Create an iterator of value/key pairs
    if isinstance(indexable, collections_abc.Mapping):
        vk_iter = ((v, k) for k, v in indexable.items())
    else:
        vk_iter = ((v, k) for k, v in enumerate(indexable))
    # Sort by values and extract the indices
    if key is None:
        indices = [k for v, k in sorted(vk_iter, reverse=reverse)]
    else:
        # If key is provided, call it using the value as input
        indices = [k for v, k in sorted(vk_iter, key=lambda vk: key(vk[0]),
                                        reverse=reverse)]
    return indices


def argmax(indexable, key=None):
    """
    Returns index / key of the item with the largest value.

    This is similar to :func:`numpy.argmax`, but it is written in pure python
    and works on both lists and dictionaries.

    Args:
        indexable (Iterable[VT] | Mapping[KT, VT]): indexable to sort by

        key (Callable[[VT], Any] | None):
            If specified, customizes the ordering of the indexable

    Returns:
        int | KT: the index of the item with the maximum value.

    Example:
        >>> import ubelt as ub
        >>> assert ub.argmax({'a': 3, 'b': 2, 'c': 100}) == 'c'
        >>> assert ub.argmax(['a', 'c', 'b', 'z', 'f']) == 3
        >>> assert ub.argmax([[0, 1], [2, 3, 4], [5]], key=len) == 1
        >>> assert ub.argmax({'a': 3, 'b': 2, 3: 100, 4: 4}) == 3
        >>> assert ub.argmax(iter(['a', 'c', 'b', 'z', 'f'])) == 3
    """
    if key is None and isinstance(indexable, collections_abc.Mapping):
        return max(indexable.items(), key=operator.itemgetter(1))[0]
    elif hasattr(indexable, 'index'):
        if key is None:
            return indexable.index(max(indexable))
        else:
            return indexable.index(max(indexable, key=key))
    else:
        # less efficient, but catch all solution
        return argsort(indexable, key=key)[-1]


def argmin(indexable, key=None):
    """
    Returns index / key of the item with the smallest value.

    This is similar to :func:`numpy.argmin`, but it is written in pure python
    and works on both lists and dictionaries.

    Args:
        indexable (Iterable[VT] | Mapping[KT, VT]): indexable to sort by

        key (Callable[[VT], VT] | None):
            If specified, customizes the ordering of the indexable.

    Returns:
        int | KT: the index of the item with the minimum value.

    Example:
        >>> import ubelt as ub
        >>> assert ub.argmin({'a': 3, 'b': 2, 'c': 100}) == 'b'
        >>> assert ub.argmin(['a', 'c', 'b', 'z', 'f']) == 0
        >>> assert ub.argmin([[0, 1], [2, 3, 4], [5]], key=len) == 2
        >>> assert ub.argmin({'a': 3, 'b': 2, 3: 100, 4: 4}) == 'b'
        >>> assert ub.argmin(iter(['a', 'c', 'A', 'z', 'f'])) == 2
    """
    if key is None and isinstance(indexable, collections_abc.Mapping):
        return min(indexable.items(), key=operator.itemgetter(1))[0]
    elif hasattr(indexable, 'index'):
        if key is None:
            return indexable.index(min(indexable))
        else:
            return indexable.index(min(indexable, key=key))
    else:
        # less efficient, but catch all solution
        return argsort(indexable, key=key)[0]


def peek(iterable, default=util_const.NoParam):
    """
    Look at the first item of an iterable. If the input is an iterator, then
    the next element is exhausted (i.e. a pop operation).

    Args:
        iterable (Iterable[T]): an iterable
        default (T): default item to return if the iterable is empty,
            otherwise a StopIteration error is raised

    Returns:
        T: item - the first item of ordered sequence, a popped item from an
                 iterator, or an arbitrary item from an unordered collection.

    Notes:
        Similar to :func:`more_itertools.peekable`

    Example:
        >>> import ubelt as ub
        >>> data = [0, 1, 2]
        >>> ub.peek(data)
        0
        >>> iterator = iter(data)
        >>> print(ub.peek(iterator))
        0
        >>> print(ub.peek(iterator))
        1
        >>> print(ub.peek(iterator))
        2
        >>> ub.peek(range(3))
        0
        >>> ub.peek([], 3)
        3
    """
    if default is util_const.NoParam:
        return next(iter(iterable))
    else:
        return next(iter(iterable), default)


# Stubs for potential future object oriented wrappers
class IterableMixin:
    """

    """
    unique = unique

    # chunks = chunks
    histogram = util_dict.dict_hist
    duplicates = util_dict.find_duplicates
    group = util_dict.group_items

    def chunks(self, size=None, num=None, bordermode='none'):
        return chunks(self, chunksize=size, nchunks=num, total=len(self), bordermode=bordermode)

    # def histogram(self, weights=None, ordered=False, labels=None):
    #     util_dict.dict_hist.__doc__
    #     return util_dict.dict_hist(self, weights=weights, ordered=ordered)

    # def duplicates(self, k=2, key=None):
    #     util_dict.find_duplicates.__doc__
    #     return util_dict.find_duplicates(self, k=k, key=key)

    # def group(self, key):
    #     util_dict.group_items.__doc__
    #     return util_dict.group_items(self, key=key)


class OrderedIterableMixin(IterableMixin):
    compress = compress
    argunique = argunique
    window = iter_window


class UList(list, OrderedIterableMixin):
    """
    An extended list class that features additional helper methods.

    Example:
        >>> from ubelt.util_list import UList
        >>> self = UList()
        >>> self.append(1)
        >>> self += UList([1, 2, 3])
        >>> self += UList([5, 7])
        >>> #
        >>> print(f'unique: {list(self.unique())}')
        >>> print(f'argunique: {list(self.argunique())}')
        >>> #
        >>> print(f'chunks: {list(self.chunks(num=2))}')
        >>> print(f'chunks: {list(self.chunks(size=2))}')
        >>> #
        >>> print(f'window: {list(self.window(3))}')
        >>> #
        >>> print(f'take: {list(self.take([0, 2, 3]))}')
        >>> print(f'compress: {list(self.compress([0, 1, 0, 1]))}')
        >>> #
        >>> print(f'argsort: {self.argsort()}')
        >>> print(f'argmax: {self.argmax()}')
        >>> print(f'argmin: {self.argmin()}')
        >>> print(f'flatten: {list(UList([self, [2, 3, 3]]).flatten())}')
        >>> print(f'allsame: {self.allsame()}')
        >>> print(f'peek: {self.peek()}')
        >>> print(f'histogram: {self.histogram()}')
        >>> print(f'group: {self.group(key=lambda x: x % 2)}')
        >>> print(f'duplicates: {self.duplicates()}')
    """
    peek = peek
    take = take

    flatten = flatten

    allsame = allsame

    argsort = argsort
    argmax = argmax
    argmin = argmin


# class USet(set, IterableMixin):
#     ...


# class Set(set, IterableMixin):
#     ...



"""
This is util_time, it contains functions for handling time related code.

The :func:`timestamp` function returns an iso8601 timestamp without much fuss.

The :func:`timeparse` is the inverse of `timestamp`, and makes use of
:mod:`dateutil` if it is available.

The :class:`Timer` class is a context manager that times a block of indented
code. It includes `tic` and `toc` methods a more matlab like feel.

Timerit is gone! Use the standalone and separate module :py:mod:`timerit`.


See Also:

    :mod:`tempora` - https://github.com/jaraco/tempora - time related utility functions from Jaraco

    :mod:`pendulum` - https://github.com/sdispater/pendulum - drop in replacement for datetime

    :mod:`arrow` - https://github.com/arrow-py/arrow

    :mod:`kwutil.util_time` - https://kwutil.readthedocs.io/en/latest/auto/kwutil.util_time.html
"""
import time
import sys
from functools import lru_cache

__all__ = ['timestamp', 'timeparse', 'Timer']


@lru_cache(maxsize=None)
def _needs_workaround39103():
    """
    Depending on the system C library, either %04Y or %Y wont work.
    This is an actual Python bug:
    https://bugs.python.org/issue13305

    singer-python also had a similar issue:
    https://github.com/singer-io/singer-python/issues/86

    See Also:
        https://github.com/jaraco/tempora/blob/main/tempora/__init__.py#L59
    """
    from datetime import datetime as datetime_cls
    return len(datetime_cls(1, 1, 1).strftime('%Y')) != 4


def timestamp(datetime=None, precision=0, default_timezone='local',
              allow_dateutil=True):
    """
    Make a concise iso8601 timestamp suitable for use in filenames.

    Args:
        datetime (datetime.datetime | datetime.date | None):
            A datetime to format into a timestamp. If unspecified, the current
            local time is used. If given as a date, the time 00:00 is used.

        precision (int):
            if non-zero, adds up to 6 digits of sub-second precision.

        default_timezone (str | datetime.timezone):
            if the input does not specify a timezone, assume this one.
            Can be "local" or "utc", or a standardized code if dateutil is
            installed.

        allow_dateutil (bool):
            if True, will use dateutil to lookup the default timezone if needed

    Returns:
        str:
            The timestamp, which will always contain a date, time, and
            timezone.

    Note:
        For more info see [WikiISO8601]_, [PyStrptime]_, [PyTime]_.

    References:
        .. [WikiISO8601] https://en.wikipedia.org/wiki/ISO_8601
        .. [PyStrptime] https://docs.python.org/3/library/datetime.html#strftime-strptime-behavior
        .. [PyTime] https://docs.python.org/3/library/time.html

    Example:
        >>> import ubelt as ub
        >>> stamp = ub.timestamp()
        >>> print('stamp = {!r}'.format(stamp))
        stamp = ...-...-...T...

    Example:
        >>> import ubelt as ub
        >>> import datetime as datetime_mod
        >>> from datetime import datetime as datetime_cls
        >>> # Create a datetime object with timezone information
        >>> ast_tzinfo = datetime_mod.timezone(datetime_mod.timedelta(hours=-4), 'AST')
        >>> datetime = datetime_cls.utcfromtimestamp(123456789.123456789).replace(tzinfo=ast_tzinfo)
        >>> stamp = ub.timestamp(datetime, precision=2)
        >>> print('stamp = {!r}'.format(stamp))
        stamp = '1973-11-29T213309.12-4'

        >>> # Demo with a fractional hour timezone
        >>> act_tzinfo = datetime_mod.timezone(datetime_mod.timedelta(hours=+9.5), 'ACT')
        >>> datetime = datetime_cls.utcfromtimestamp(123456789.123456789).replace(tzinfo=act_tzinfo)
        >>> stamp = ub.timestamp(datetime, precision=2)
        >>> print('stamp = {!r}'.format(stamp))
        stamp = '1973-11-29T213309.12+0930'

        >>> # Can accept datetime or date objects with local, utc, or custom default timezones
        >>> act_tzinfo = datetime_mod.timezone(datetime_mod.timedelta(hours=+9.5), 'ACT')
        >>> datetime_utc = ub.timeparse('2020-03-05T112233', default_timezone='utc')
        >>> datetime_act = ub.timeparse('2020-03-05T112233', default_timezone=act_tzinfo)
        >>> datetime_notz = datetime_utc.replace(tzinfo=None)
        >>> date = datetime_utc.date()
        >>> stamp_utc = ub.timestamp(datetime_utc)
        >>> stamp_act = ub.timestamp(datetime_act)
        >>> stamp_date_utc = ub.timestamp(date, default_timezone='utc')
        >>> print(f'stamp_utc      = {stamp_utc}')
        >>> print(f'stamp_act      = {stamp_act}')
        >>> print(f'stamp_date_utc = {stamp_date_utc}')
        stamp_utc      = 2020-03-05T112233+0
        stamp_act      = 2020-03-05T112233+0930
        stamp_date_utc = 2020-03-05T000000+0

    Example:
        >>> # xdoctest: +REQUIRES(module:dateutil)
        >>> # Make sure we are compatible with dateutil
        >>> import ubelt as ub
        >>> from dateutil.tz import tzlocal
        >>> import datetime as datetime_mod
        >>> from datetime import datetime as datetime_cls
        >>> tz_act = datetime_mod.timezone(datetime_mod.timedelta(hours=+9.5), 'ACT')
        >>> tzinfo_list = [
        >>>     tz_act,
        >>>     datetime_mod.timezone(datetime_mod.timedelta(hours=-4), 'AST'),
        >>>     datetime_mod.timezone(datetime_mod.timedelta(hours=0), 'UTC'),
        >>>     datetime_mod.timezone.utc,
        >>>     None,
        >>>     tzlocal()
        >>> ]
        >>> # Note: there is a win32 bug here
        >>> # https://bugs.python.org/issue37 that means we cant use
        >>> # dates close to the epoch
        >>> datetime_list = [
        >>>     datetime_cls.utcfromtimestamp(123456789.123456789 + 315360000),
        >>>     datetime_cls.utcfromtimestamp(0 + 315360000),
        >>> ]
        >>> basis = {
        >>>     'precision': [0, 3, 9],
        >>>     'tzinfo': tzinfo_list,
        >>>     'datetime': datetime_list,
        >>>     'default_timezone': ['local', 'utc', tz_act],
        >>> }
        >>> for params in ub.named_product(basis):
        >>>     dtime = params['datetime'].replace(tzinfo=params['tzinfo'])
        >>>     precision = params.get('precision', 0)
        >>>     stamp = ub.timestamp(datetime=dtime, precision=precision)
        >>>     recon = ub.timeparse(stamp)
        >>>     alt = recon.strftime('%Y-%m-%dT%H%M%S.%f%z')
        >>>     print('---')
        >>>     print('params = {}'.format(ub.repr2(params, nl=1)))
        >>>     print(f'dtime={dtime}')
        >>>     print(f'stamp={stamp}')
        >>>     print(f'recon={recon}')
        >>>     print(f'alt  ={alt}')
        >>>     shift = 10 ** precision
        >>>     a = int(dtime.timestamp() * shift)
        >>>     b = int(recon.timestamp() * shift)
        >>>     assert a == b, f'{a} != {b}'
    """
    import datetime as datetime_mod
    from datetime import datetime as datetime_cls

    datetime_obj = datetime
    offset_seconds = None

    # datetime inherits from date. Strange, so we cant use this. See:
    # https://github.com/python/typeshed/issues/4802
    if isinstance(datetime_obj, datetime_mod.date) and not isinstance(datetime_obj, datetime_cls):
        # Coerce a date to datetime.datetime
        datetime_obj = datetime_cls.combine(datetime_obj, datetime_cls.min.time())

    if datetime_obj is None or datetime_obj.tzinfo is None:
        # In either case, we need to construct a timezone object
        tzinfo = _timezone_coerce(default_timezone,
                                  allow_dateutil=allow_dateutil)
        # If datetime_obj is unspecified, create a timezone aware now object
        if datetime_obj is None:
            datetime_obj = datetime_cls.now(tzinfo)
    else:
        tzinfo = datetime_obj.tzinfo

    # the arg to utcoffset is confusing
    offset_seconds = tzinfo.utcoffset(datetime_obj).total_seconds()
    # offset_seconds = tzinfo.utcoffset(None).total_seconds()

    seconds_per_hour = 3600
    tz_hour, tz_remain = divmod(offset_seconds, seconds_per_hour)
    tz_hour = int(tz_hour)
    if tz_remain:
        seconds_per_minute = 60
        tz_min = int(tz_remain // seconds_per_minute)
        utc_offset = '{:+03d}{:02d}'.format(tz_hour, tz_min)
    else:
        utc_offset = str(tz_hour) if tz_hour < 0 else '+' + str(tz_hour)
    if precision > 0:
        fprecision = 6  # microseconds are padded to 6 decimals
        # NOTE: The time.strftime and datetime.datetime.strftime methods
        # seem to work differently. The former does not support %f
        if _needs_workaround39103():  # nocover
            local_stamp = datetime_obj.strftime('%04Y-%m-%dT%H%M%S.%f')
        else:  # nocover
            local_stamp = datetime_obj.strftime('%Y-%m-%dT%H%M%S.%f')
        ms_offset = len(local_stamp) - max(0, fprecision - precision)
        local_stamp = local_stamp[:ms_offset]
    else:
        if _needs_workaround39103():  # nocover
            local_stamp = datetime_obj.strftime('%04Y-%m-%dT%H%M%S')
        else:  # nocover
            local_stamp = datetime_obj.strftime('%Y-%m-%dT%H%M%S')
    stamp = local_stamp + utc_offset
    return stamp


def timeparse(stamp, default_timezone='local', allow_dateutil=True):
    """
    Create a :class:`datetime.datetime` object from a string timestamp.

    Without any extra dependencies this will parse the output of
    :func:`ubelt.util_time.timestamp` into a datetime object. In the case
    where the format differs, :func:`dateutil.parser.parse` will be used
    if the :py:mod:`python-dateutil` package is installed.

    Args:
        stamp (str):
            a string encoded timestamp

        default_timezone (str):
            if the input does not specify a timezone, assume this one.
            Can be "local" or "utc".

        allow_dateutil (bool):
            if False we only use the minimal parsing and do not allow a
            fallback to dateutil.

    Returns:
        datetime.datetime: the parsed datetime

    Raises:
        ValueError: if if parsing fails.

    TODO:
        - [ ] Allow defaulting to local or utm timezone (currently default is local)

    Example:
        >>> import ubelt as ub
        >>> # Demonstrate a round trip of timestamp and timeparse
        >>> stamp = ub.timestamp()
        >>> datetime = ub.timeparse(stamp)
        >>> assert ub.timestamp(datetime) == stamp
        >>> # Round trip with precision
        >>> stamp = ub.timestamp(precision=4)
        >>> datetime = ub.timeparse(stamp)
        >>> assert ub.timestamp(datetime, precision=4) == stamp

    Example:
        >>> import ubelt as ub
        >>> # We should always be able to parse these
        >>> good_stamps = [
        >>>     '2000-11-22',
        >>>     '2000-11-22T111111.44444Z',
        >>>     '2000-11-22T111111.44444+5',
        >>>     '2000-11-22T111111.44444-05',
        >>>     '2000-11-22T111111.44444-0500',
        >>>     '2000-11-22T111111.44444+0530',
        >>>     '2000-11-22T111111Z',
        >>>     '2000-11-22T111111+5',
        >>>     '2000-11-22T111111+0530',
        >>> ]
        >>> for stamp in good_stamps:
        >>>     print(f'----')
        >>>     print(f'stamp={stamp}')
        >>>     result = ub.timeparse(stamp, allow_dateutil=0)
        >>>     print(f'result={result!r}')
        >>>     recon = ub.timestamp(result)
        >>>     print(f'recon={recon}')

    Example:
        >>> import ubelt as ub
        >>> # We require dateutil to handle these types of stamps
        >>> import pytest
        >>> conditional_stamps = [
        >>>         '2000-01-02T11:23:58.12345+5:30',
        >>>         '09/25/2003',
        >>>         'Thu Sep 25 10:36:28 2003',
        >>> ]
        >>> for stamp in conditional_stamps:
        >>>     with pytest.raises(ValueError):
        >>>         result = ub.timeparse(stamp, allow_dateutil=False)
        >>> have_dateutil = bool(ub.modname_to_modpath('dateutil'))
        >>> if have_dateutil:
        >>>     for stamp in conditional_stamps:
        >>>         result = ub.timeparse(stamp)

    Ignore:
        import timerit
        ti = timerit.Timerit(1000, 10)
        ti.reset('non-standard dateutil.parse').call(lambda: dateutil.parser.parse('2000-01-02T112358.12345+5'))
        ti.reset('non-standard ubelt.timeparse').call(lambda: ub.timeparse('2000-01-02T112358.12345+5'))
        ti.reset('standard dateutil.parse').call(lambda: dateutil.parser.parse('2000-01-02T112358.12345+0500'))
        ti.reset('standard dateutil.isoparse').call(lambda: dateutil.parser.isoparse('2000-01-02T112358.12345+0500'))
        ti.reset('standard ubelt.timeparse').call(lambda: ub.timeparse('2000-01-02T112358.12345+0500'))
        ti.reset('standard datetime_cls.strptime').call(lambda: datetime_cls.strptime('2000-01-02T112358.12345+0500', '%Y-%m-%dT%H%M%S.%f%z'))
    """
    from datetime import datetime as datetime_cls
    datetime_obj = None
    # Check if we might have a minimal format
    maybe_minimal = (
        len(stamp) >= 17 and 'T' in stamp[10:]
    )
    fixed_stamp = stamp
    if maybe_minimal:
        # Note by default %z only handles the format `[+-]HHMM(SS(.ffffff))`
        # this means we have to handle the case where `[+-]HH` is given.
        # We do this by checking the offset and padding it to at least the
        # `[+-]HHMM` format
        date_part, timetz_part = stamp.split('T', 1)
        if '-' in timetz_part[6:]:
            time_part, sign, tz_part = timetz_part.partition('-')
        elif '+' in timetz_part[6:]:
            time_part, sign, tz_part = timetz_part.partition('+')
        else:
            # In 3.7 a Z suffix is handled correctly
            # For 3.6 compatibility, replace Z with +0000
            if timetz_part.endswith('Z'):
                time_part = timetz_part[:-1]
                sign = '+'
                tz_part = '0000'
            else:
                tz_part = None

        if tz_part is not None:
            if len(tz_part) == 1:
                tz_part = '0{}00'.format(tz_part)
            elif len(tz_part) == 2:
                tz_part = '{}00'.format(tz_part)
            fixed_stamp = ''.join([date_part, 'T', time_part, sign, tz_part])

    if len(stamp) == 10:
        try:
            fmt =  '%Y-%m-%d'
            datetime_obj = datetime_cls.strptime(fixed_stamp, fmt)
        except ValueError:
            pass

    if maybe_minimal and datetime_obj is None:
        minimal_formats = [
            '%Y-%m-%dT%H%M%S%z',
            '%Y-%m-%dT%H%M%S',
            '%Y-%m-%dT%H%M%S.%f%z',
            '%Y-%m-%dT%H%M%S.%f',
        ]
        for fmt in minimal_formats:
            try:
                datetime_obj = datetime_cls.strptime(fixed_stamp, fmt)
            except ValueError:
                pass
            else:
                break

    if datetime_obj is None:
        # Our minimal logic did not work, can we use dateutil?
        if not allow_dateutil:
            raise ValueError((
                'Cannot parse timestamp. '
                'Unknown string format: {!r}, and '
                'dateutil is not allowed').format(stamp))
        else:
            try:
                from dateutil.parser import parse as du_parse
            except (ModuleNotFoundError, ImportError):  # nocover
                raise ValueError((
                    'Cannot parse timestamp. '
                    'Unknown string format: {!r}, and '
                    'dateutil is not installed').format(stamp)) from None
            else:  # nocover
                datetime_obj = du_parse(stamp)

    if datetime_obj.tzinfo is None:
        # Timezone is unspecified, need to construct the default one.
        tzinfo = _timezone_coerce(default_timezone,
                                  allow_dateutil=allow_dateutil)
        datetime_obj = datetime_obj.replace(tzinfo=tzinfo)

    return datetime_obj


def _timezone_coerce(tzinfo, allow_dateutil=True):
    """
    Ensure output it a timezone instance.

    Example:
        >>> import pytest
        >>> from ubelt.util_time import *  # NOQA
        >>> from ubelt.util_time import _timezone_coerce
        >>> results = []
        >>> write = results.append
        >>> tzinfo = _timezone_coerce('utc', allow_dateutil=0)
        >>> print(f'tzinfo={tzinfo}')
        tzinfo=UTC

    Example:
        >>> # xdoctest: +REQUIRES(module:dateutil)
        >>> import pytest
        >>> results = []
        >>> write = results.append
        >>> write(_timezone_coerce('utc'))
        >>> write(_timezone_coerce('GMT'))
        >>> write(_timezone_coerce('EST'))
        >>> write(_timezone_coerce('HST'))
        >>> import datetime as datetime_mod
        >>> dt = datetime_mod.datetime.now()
        >>> for tzinfo in results:
        >>>     print(f'tzoffset={tzinfo.utcoffset(dt).total_seconds()}')
        tzoffset=0.0
        tzoffset=0.0
        tzoffset=-18000.0
        tzoffset=-36000.0

    Example:
        >>> import pytest
        >>> from ubelt.util_time import *  # NOQA
        >>> from ubelt.util_time import _timezone_coerce
        >>> with pytest.raises(ValueError):
        ...     _timezone_coerce('GMT', allow_dateutil=0)
        >>> with pytest.raises(TypeError):
        ...     _timezone_coerce(object(), allow_dateutil=0)
        >>> # xdoctest: +REQUIRES(module:dateutil)
        >>> with pytest.raises(KeyError):
        ...     _timezone_coerce('NotATimezone', allow_dateutil=1)

    Example:
        >>> import pytest
        >>> from ubelt.util_time import *  # NOQA
        >>> from ubelt.util_time import _timezone_coerce
        >>> import time
        >>> tz1 = _timezone_coerce('local', allow_dateutil=0)
        >>> tz2 = _timezone_coerce('local', allow_dateutil=1)
        >>> sec1 = tz1.utcoffset(None).total_seconds()
        >>> sec2 = tz2.utcoffset(None).total_seconds()
        >>> assert sec1 == sec2 == -time.timezone
    """
    import datetime as datetime_mod
    if isinstance(tzinfo, str):
        if tzinfo == 'local':
            # Note: the local timezone time.timezone is negated
            _delta = datetime_mod.timedelta(seconds=-time.timezone)
            out_tzinfo = datetime_mod.timezone(_delta)
        elif tzinfo == 'utc':
            out_tzinfo = datetime_mod.timezone.utc
        else:
            if allow_dateutil:
                from dateutil import tz as tz_mod
                out_tzinfo = tz_mod.gettz(tzinfo)
                if out_tzinfo is None:
                    raise KeyError(tzinfo)
            else:
                raise ValueError((
                    'Unrecognized timezone: {!r}, and '
                    'dateutil is not allowed').format(tzinfo))
    elif isinstance(tzinfo, datetime_mod.timezone):
        out_tzinfo = tzinfo
    else:
        raise TypeError(
            'Unknown type: {!r} for tzinfo'.format(
                type(tzinfo))
        )
    return out_tzinfo


class Timer(object):
    """
    Measures time elapsed between a start and end point. Can be used as a
    with-statement context manager, or using the tic/toc api.

    Attributes:
        elapsed (float): number of seconds measured by the context manager
        tstart (float): time of last `tic` reported by `self._time()`
        write (Callable): function used to write
        flush (Callable): function used to flush

    Example:
        >>> # Create and start the timer using the context manager
        >>> import math
        >>> import ubelt as ub
        >>> timer = ub.Timer('Timer test!', verbose=1)
        >>> with timer:
        >>>     math.factorial(10)
        >>> assert timer.elapsed > 0
        tic('Timer test!')
        ...toc('Timer test!')=...

    Example:
        >>> # Create and start the timer using the tic/toc interface
        >>> import ubelt as ub
        >>> timer = ub.Timer().tic()
        >>> elapsed1 = timer.toc()
        >>> elapsed2 = timer.toc()
        >>> elapsed3 = timer.toc()
        >>> assert elapsed1 <= elapsed2
        >>> assert elapsed2 <= elapsed3

    Example:
        >>> # In Python 3.7+ nanosecond resolution can be enabled
        >>> import ubelt as ub
        >>> import sys
        >>> if sys.version_info[0:2] <= (3, 6):
        >>>     import pytest
        >>>     pytest.skip()
        >>> # xdoctest +REQUIRES(Python>=3.7)  # fixme directive doesnt exist yet
        >>> timer = ub.Timer(label='perf_counter_ns', ns=True).tic()
        >>> elapsed1 = timer.toc()
        >>> elapsed2 = timer.toc()
        >>> assert elapsed1 <= elapsed2
        >>> assert isinstance(elapsed1, int)
    """
    _default_time = time.perf_counter

    def __init__(self, label='', verbose=None, newline=True, ns=False):
        """
        Args:
            label (str):
                identifier for printing. Default to ''.

            verbose (int | None):
                verbosity flag, defaults to True if label is given, otherwise 0.

            newline (bool):
                if False and verbose, print tic and toc on the same line.
                Defaults to True.

            ns (bool):
                if True, a nano-second resolution timer to avoid precision loss
                caused by the float type. Defaults to False.
        """
        if verbose is None:
            verbose = bool(label)
        self.label = label
        self.verbose = verbose
        self.newline = newline
        self.tstart = -1
        self.elapsed = -1
        self.write = sys.stdout.write
        self.flush = sys.stdout.flush
        self.ns = ns
        if self.ns:
            self._time = time.perf_counter_ns
        else:
            self._time = self._default_time

    def tic(self):
        """
        starts the timer

        Returns:
            Timer: self
        """
        if self.verbose:
            self.flush()
            self.write('\ntic(%r)' % self.label)
            if self.newline:
                self.write('\n')
            self.flush()
        self.tstart = self._time()
        return self

    def toc(self):
        """
        stops the timer

        Returns:
            float | int: number of second or nanoseconds
        """
        elapsed = self._time() - self.tstart
        if self.verbose:
            if self.ns:
                self.write('...toc(%r)=%.4fs\n' % (self.label, elapsed / 1e9))
            else:
                self.write('...toc(%r)=%.4fs\n' % (self.label, elapsed))
            self.flush()
        return elapsed

    def __enter__(self):
        """
        Returns:
            Timer: self
        """
        self.tic()
        return self

    def __exit__(self, ex_type, ex_value, ex_traceback):
        """
        Args:
            ex_type (Type[BaseException] | None):
            ex_value (BaseException | None):
            ex_traceback (TracebackType | None):

        Returns:
            bool | None
        """
        self.elapsed = self.toc()
        if ex_traceback is not None:
            return False


# class Time:
#     """
#     Stub for potential future time object
#     """
#     def __init__(cls, datetime):
#         ...
#     @classmethod
#     def coerce(cls, data):
#         ...
#     @classmethod
#     def parse(cls, stamp):
#         ...
#
# class TimeDelta:
#     ...



import sys
from typing import Iterable
from typing import Type
from typing import Dict
from typing import Callable
from typing import List
from typing import Any
from typing import Optional
from ubelt.util_const import NoParam
from ubelt.util_const import NoParamType
from collections import OrderedDict
from typing import Mapping
from typing import Set
from typing import Tuple
from _typeshed import SupportsKeysAndGetItem
from collections import OrderedDict, defaultdict
from collections.abc import Generator
from typing import Any, TypeVar

VT = TypeVar("VT")
T = TypeVar("T")
KT = TypeVar("KT")
odict = OrderedDict
ddict = defaultdict
DictBase = OrderedDict if sys.version_info[0:2] <= (3, 6) else dict


def dzip(items1: Iterable[KT],
         items2: Iterable[VT],
         cls: Type[dict] = dict) -> Dict[KT, VT]:
    ...


def group_items(items: Iterable[VT],
                key: Iterable[KT] | Callable[[VT], KT]) -> dict[KT, List[VT]]:
    ...


def dict_hist(items: Iterable[T],
              weights: Iterable[float] | None = None,
              ordered: bool = False,
              labels: Iterable[T] | None = None) -> dict[T, int]:
    ...


def find_duplicates(
        items: Iterable[T],
        k: int = 2,
        key: Callable[[T], Any] | None = None) -> dict[T, List[int]]:
    ...


def dict_subset(dict_: Dict[KT, VT],
                keys: Iterable[KT],
                default: Optional[object] | NoParamType = NoParam,
                cls: Type[Dict] = OrderedDict) -> Dict[KT, VT]:
    ...


def dict_union(*args: List[Dict]) -> Dict | OrderedDict:
    ...


def dict_diff(
    *args: List[Dict[KT, VT] | Iterable[KT]]
) -> Dict[KT, VT] | OrderedDict[KT, VT]:
    ...


def dict_isect(
    *args: List[Dict[KT, VT] | Iterable[KT]]
) -> Dict[KT, VT] | OrderedDict[KT, VT]:
    ...


def map_values(func: Callable[[VT], T] | Mapping[VT, T],
               dict_: Dict[KT, VT],
               cls: type | None = None) -> Dict[KT, T]:
    ...


map_vals = map_values


def map_keys(func: Callable[[KT], T] | Mapping[KT, T],
             dict_: Dict[KT, VT],
             cls: type | None = None) -> Dict[T, VT]:
    ...


def sorted_values(dict_: Dict[KT, VT],
                  key: Callable[[VT], Any] | None = None,
                  reverse: bool = False,
                  cls: type = OrderedDict) -> OrderedDict[KT, VT]:
    ...


sorted_vals = sorted_values


def sorted_keys(dict_: Dict[KT, VT],
                key: Callable[[KT], Any] | None = None,
                reverse: bool = False,
                cls: type = OrderedDict) -> OrderedDict[KT, VT]:
    ...


def invert_dict(dict_: Dict[KT, VT],
                unique_vals: bool = True,
                cls: type | None = None) -> Dict[VT, KT] | Dict[VT, Set[KT]]:
    ...


def named_product(
        _: Dict[str, List[VT]] | None = None,
        **basis: Dict[str, List[VT]]) -> Generator[Dict[str, VT], None, None]:
    ...


def varied_values(longform: List[Dict[KT, VT]],
                  min_variations: int = 0,
                  default: VT | NoParamType = NoParam) -> Dict[KT, List[VT]]:
    ...


class SetDict(dict):

    def copy(self):
        ...

    def __or__(
        self,
        other: SupportsKeysAndGetItem[Any, Any] | Iterable[Tuple[Any, Any]]
    ) -> SetDict:
        ...

    def __and__(self, other: Mapping | Iterable) -> SetDict:
        ...

    def __sub__(self, other: Mapping | Iterable) -> SetDict:
        ...

    def __xor__(self, other: Mapping) -> SetDict:
        ...

    def __ror__(self, other: Mapping) -> dict:
        ...

    def __rand__(self, other: Mapping) -> dict:
        ...

    def __rsub__(self, other: Mapping) -> dict:
        ...

    def __rxor__(self, other: Mapping) -> dict:
        ...

    def __ior__(
        self,
        other: SupportsKeysAndGetItem[Any, Any] | Iterable[Tuple[Any, Any]]
    ) -> SetDict:
        ...

    def __iand__(self, other: Mapping | Iterable):
        ...

    def __isub__(self, other: Mapping | Iterable):
        ...

    def __ixor__(self, other: Mapping):
        ...

    def union(self,
              *others,
              cls: type | None = None,
              merge: None | Callable = None) -> dict:
        ...

    def intersection(self,
                     *others,
                     cls: type | None = None,
                     merge: None | Callable = None) -> dict:
        ...

    def difference(self,
                   *others,
                   cls: type | None = None,
                   merge: None | Callable = None) -> dict:
        ...

    def symmetric_difference(self,
                             *others,
                             cls: type | None = None,
                             merge: None | Callable = None) -> dict:
        ...


sdict = SetDict


class UDict(SetDict):

    def subdict(self,
                keys: Iterable[KT],
                default: Optional[object] | NoParamType = NoParam):
        ...

    def take(
        self,
        keys: Iterable[KT],
        default: Optional[object] | NoParamType = NoParam
    ) -> Generator[VT, None, None]:
        ...

    def invert(self,
               unique_vals: bool = True) -> Dict[VT, KT] | Dict[VT, Set[KT]]:
        ...

    def map_keys(self,
                 func: Callable[[VT], T] | Mapping[VT, T]) -> Dict[KT, T]:
        ...

    def map_values(self,
                   func: Callable[[VT], T] | Mapping[VT, T]) -> Dict[KT, T]:
        ...

    def sorted_keys(self,
                    key: Callable[[KT], Any] | None = None,
                    reverse: bool = False) -> OrderedDict[KT, VT]:
        ...

    def sorted_values(self,
                      key: Callable[[VT], Any] | None = None,
                      reverse: bool = False) -> OrderedDict[KT, VT]:
        ...

    def peek_key(self, default: KT | NoParamType = NoParam) -> KT:
        ...

    def peek_value(self, default: VT | NoParamType = NoParam) -> VT:
        ...


class AutoDict(UDict):

    def __getitem__(self, key: KT) -> VT | AutoDict:
        ...

    def to_dict(self) -> dict:
        ...


AutoOrderedDict = AutoDict
udict = UDict



from os.path import exists, join
import ubelt as ub
# DEBUG_PATH = ub.Path.home().name == 'joncrall'


def test_pathlib_compatability():
    import pathlib
    base = pathlib.Path(ub.Path.appdir('ubelt').ensuredir())
    dpath = base.joinpath('test_pathlib_mkdir')

    # ensuredir
    ub.delete(dpath)
    assert not dpath.exists()
    got = ub.ensuredir(dpath)
    assert got.exists()

    # shrinkuser
    assert ub.shrinkuser(base).startswith('~')

    assert ub.augpath(base, prefix='foo').endswith('fooubelt')
    assert not ub.expandpath(base).startswith('~')


def test_tempdir():
    import pytest
    with pytest.warns(DeprecationWarning):
        temp = ub.TempDir()
    assert temp.dpath is None
    temp.ensure()
    assert exists(temp.dpath)
    # Double ensure for coverage
    temp.ensure()
    assert exists(temp.dpath)

    dpath = temp.dpath
    temp.cleanup()
    assert not exists(dpath)
    assert temp.dpath is None


def test_augpath_identity():
    assert ub.augpath('foo') == 'foo'
    assert ub.augpath('foo/bar') == join('foo', 'bar')
    assert ub.augpath('') == ''


def test_augpath_dpath():
    assert ub.augpath('foo', dpath='bar') == join('bar', 'foo')
    assert ub.augpath('foo/bar', dpath='baz') == join('baz', 'bar')
    assert ub.augpath('', dpath='bar').startswith('bar')


def test_ensuredir_recreate():
    import pytest
    base = ub.Path.appdir('ubelt/tests').ensuredir()
    folder = join(base, 'foo')
    member = join(folder, 'bar')
    with pytest.warns(DeprecationWarning):
        ub.ensuredir(folder, recreate=True)
    ub.ensuredir(member)
    assert exists(member)
    with pytest.warns(DeprecationWarning):
        ub.ensuredir(folder, recreate=True)
    assert not exists(member)


def test_ensuredir_verbosity():
    base = ub.Path.appdir('ubelt/tests').ensuredir()

    with ub.CaptureStdout() as cap:
        ub.ensuredir(join(base, 'foo'), verbose=0)
    assert cap.text == ''
    # None defaults to verbose=0
    with ub.CaptureStdout() as cap:
        ub.ensuredir((base, 'foo'), verbose=None)
    assert cap.text == ''

    ub.delete(join(base, 'foo'))
    with ub.CaptureStdout() as cap:
        ub.ensuredir(join(base, 'foo'), verbose=1)
    assert 'creating' in cap.text
    with ub.CaptureStdout() as cap:
        ub.ensuredir(join(base, 'foo'), verbose=1)
    assert 'existing' in cap.text


def demo_nested_paths(dpath, nfiles=2, ndirs=1, depth=0):
    for idx in range(nfiles):
        (dpath / f'file_{idx}.txt').write_text(f'hello world idx={idx} depth={depth}')

    subdirs = []
    for idx in range(ndirs):
        subdir = (dpath / f'subdir_{idx}').ensuredir()
        subdirs.append(subdir)

    if depth > 0:
        for subdir in subdirs:
            demo_nested_paths(subdir, nfiles=nfiles, ndirs=ndirs,
                                    depth=depth - 1)


def relative_contents(dpath):
    return [p.relative_to(dpath) for p in sorted(dpath.glob('**'), key=str)]


def test_copy_directory_cases():
    """
    Ignore:

        cases = [
            {'dst': '{}'},
        ]

    """
    import pytest
    import ubelt as ub
    base = ub.Path.appdir('ubelt/tests/path/copy_move').delete().ensuredir()

    root1 = (base / 'root1').ensuredir()
    root2 = (base / 'root2').ensuredir()
    paths = {
        'empty': root1 / 'empty',
        'shallow': root1 / 'shallow',
        'deep': root1 / 'deep',
    }
    for d in paths.values():
        d.ensuredir()
    demo_nested_paths(paths['shallow'])
    demo_nested_paths(paths['deep'], depth=3)

    # Instead you can always expect <dst>/<contents> to be the same as
    # <src>/<contents>.
    for key, src in paths.items():
        for meta in ['stats', 'mode', None]:
            kwargs = {
                'meta': meta
            }
            root2.delete().ensuredir()
            # Because root2 exists we error if overwrite if False
            with pytest.raises(FileExistsError):
                src.copy(root2, **kwargs)
            # When overwrite is True,
            src.copy(root2, overwrite=True, **kwargs)
            relative_contents(root2)
            contents1 = relative_contents(src)
            contents2 = relative_contents(root2)
            assert contents1 == contents2

            # We can copy to a directory that doesn't exist
            root2.delete().ensuredir()
            new_dpath = src.copy(root2 / src.name, **kwargs)
            assert new_dpath.name == src.name
            contents2 = relative_contents(new_dpath)
            # But we can't do it again
            with pytest.raises(FileExistsError):
                src.copy(root2 / src.name, **kwargs)
            assert contents2 == relative_contents(new_dpath)
            # Unless overwrite is True
            new_dpath = src.copy(root2 / src.name, overwrite=True, **kwargs)
            # And in all cases the contents should be unchanged
            assert contents2 == relative_contents(new_dpath)

            # Test copy src into root2/sub1/sub2 when root/sub1 does not exist
            root2.delete().ensuredir()
            dst = root2 / 'sub1/sub2'
            new_dpath = src.copy(dst, **kwargs)
            assert new_dpath.name == 'sub2'
            # Unlike cp, Path.copy will create the intermediate directories
            assert contents1 == relative_contents(new_dpath)

        if 0 and ub.LINUX:
            """
            In all cases we have a folder
                <src> = <parent1>/<name>

                with members

                <parent1>/<name>/<contents>

                and

                <dst> = <parent2>/<dname>
            """
            verbose = 2
            """
            Case:
                copy ``<parent>/<name>`` into ``<dst>``
                and  ``<dst>/<name>`` does not exist,
                then cp will result in ``<dst>/<name>/<contents>``

                THEN

                copy ``<parent1>/<name>`` into ``<dst>``
                and  ``<dst>/<name>`` exists,
                then cp will result in ``<dst>/<name>/<contents>``

            CP recognizes that ``<dst>/<name>`` does not exist and makes a new
            directory <dst>/<name> to correspond to to the <src>

            THEN

            CP recognizes that ``<dst>/<name>`` does exist assumes
            that should correspond to <src>
            """
            root2.delete().ensuredir()
            dst = root2
            ub.cmd(f"cp -rv {src} {dst}", verbose=verbose)
            contents1 = relative_contents(src)
            contents2 = relative_contents(root2)
            assert len(contents1) == (len(contents2) - 1)

            ub.cmd(f"cp -rv {src} {dst}", verbose=verbose)
            contents1 = relative_contents(src)
            contents2 = relative_contents(root2)
            assert len(contents1) == (len(contents2) - 1)

            """
            Case:
                copy ``<parent1>/<name>`` into ``<parent2>/<name2>``
                and  ``<parent2>/<name2>`` does not exist,
                then cp will result in ``<dst>/<name2>/<contents>``

                THEN

                copy ``<parent1>/<name>`` into ``<parent2>/<name2>``
                and  ``<parent2>/<name2>`` does exist,
                then cp will result in ``<parent2>/<name2>/<name>/<contents>``

            Because <name2> in <parent> does not exist, it assumes you want to
            effectively *change the name* of your folder.

            THEN

            This is the weird case that Path.copy avoids by not pretending that
            you can use the directory name from the source implicitly in the
            destination.
            """
            root2.delete().ensuredir()
            name2 = f'{src.name}2'
            dst = root2 / name2
            ub.cmd(f"cp -rv {src} {dst}", verbose=verbose)
            contents1 = relative_contents(src)
            contents2 = relative_contents(root2)
            assert len(contents1) == (len(contents2) - 1)

            dst = root2 / name2
            ub.cmd(f"cp -rv {src} {dst}", verbose=verbose)
            contents1 = relative_contents(src)
            contents2 = relative_contents(root2)
            assert len(contents1) * 2 == (len(contents2) - 1)

            """
            Case:
                copy ``<parent>/<name>`` into ``<parent2>/<sub1>/<sub2>``
                and  ``<parent2>/<sub1>`` does not exist,
                then cp will error because it wont create intermediate
                directories
            """
            root2.delete().ensuredir()
            dst = root2 / 'sub1/sub2'
            info = ub.cmd(f"cp -rv {src} {dst}", verbose=verbose)
            assert info['ret'] == 1
            contents2 = relative_contents(root2)
            assert len(contents2) == 1


def test_move_directory_cases():
    """
    Ignore:

        cases = [
            {'dst': '{}'},
        ]

    """
    import pytest
    import ubelt as ub
    base = ub.Path.appdir('ubelt/tests/path/move').delete().ensuredir()

    root1 = (base / 'root1').ensuredir()
    root2 = (base / 'root2').ensuredir()
    paths = {
        'empty': root1 / 'empty',
        'shallow': root1 / 'shallow',
        'deep': root1 / 'deep',
    }
    for d in paths.values():
        d.ensuredir()

    # Instead you can always expect <dst>/<contents> to be the same as
    # <src>/<contents>.
    for key, src in paths.items():
        for meta in ['stats', 'mode', None]:

            # Reset original dires
            for d in paths.values():
                d.ensuredir()
            demo_nested_paths(paths['shallow'])
            demo_nested_paths(paths['deep'], depth=3)

            kwargs = {
                'meta': meta
            }
            root2.delete().ensuredir()
            # We cannot move to a file that exists
            with pytest.raises(FileExistsError):
                src.move(root2, **kwargs)

            contents1 = relative_contents(src)
            # We can move to a directory that doesn't exist
            root2.delete().ensuredir()
            new_dpath = src.move(root2 / src.name, **kwargs)
            assert new_dpath.name == src.name
            contents2 = relative_contents(new_dpath)
            assert not src.exists()
            assert contents1 == contents2

            with pytest.raises(FileExistsError):
                src.move(root2 / src.name, **kwargs)

            # Test move src into root2/sub1/sub2 when root/sub1 does not exist
            # Reset original dires
            for d in paths.values():
                d.ensuredir()
            demo_nested_paths(paths['shallow'])
            demo_nested_paths(paths['deep'], depth=3)
            root2.delete().ensuredir()
            dst = root2 / 'sub1/sub2'
            new_dpath = src.move(dst, **kwargs)
            assert new_dpath.name == 'sub2'
            # Unlike cp, Path.move will create the intermediate directories
            assert contents1 == relative_contents(new_dpath)


def test_follow_file_symlinks():
    import ubelt as ub
    root = ub.Path.appdir('ubelt', 'tests', 'path', 'copy-file-symlink').delete().ensuredir()
    fpath1 = (root / 'file').touch()
    flink1 = (root / 'flink1')
    flink1.symlink_to(fpath1)
    fcopy1 = flink1.copy(root / 'fcopy1', follow_file_symlinks=True, meta=None)
    fcopy2 = flink1.copy(root / 'fcopy2', follow_file_symlinks=False, meta=None)
    assert not fcopy1.is_symlink(), 'should have followed symlink'
    assert fcopy2.is_symlink(), 'should not have followed symlink'

    fcopy3 = flink1.copy(root / 'fcopy3', follow_file_symlinks=True, meta='mode')
    fcopy4 = flink1.copy(root / 'fcopy4', follow_file_symlinks=False, meta='mode')
    assert not fcopy3.is_symlink(), 'should have followed symlink'
    assert fcopy4.is_symlink(), 'should not have followed symlink'

    fcopy5 = flink1.copy(root / 'fcopy5', follow_file_symlinks=True, meta='stats')
    fcopy6 = flink1.copy(root / 'fcopy6', follow_file_symlinks=False, meta='stats')
    assert not fcopy5.is_symlink(), 'should have followed symlink'
    assert fcopy6.is_symlink(), 'should not have followed symlink'



from typing import Iterable
from _typeshed import SupportsWrite
from typing import List
import typing
from typing import Callable
from typing import Type
from types import TracebackType
from typing import Tuple
from _typeshed import Incomplete
from typing import NamedTuple

default_timer: Callable


class Measurement(NamedTuple):
    idx: Incomplete
    time: Incomplete


CLEAR_BEFORE: str
AT_END: str


class _TQDMCompat:

    @classmethod
    def write(cls,
              s: str,
              file: None | SupportsWrite = None,
              end: str = '\n',
              nolock: bool = False) -> None:
        ...

    desc: str | None

    def set_description(self,
                        desc: str | None = None,
                        refresh: bool = ...) -> None:
        ...

    def set_description_str(self,
                            desc: str | None = None,
                            refresh: bool = ...) -> None:
        ...

    def update(self, n: int = ...) -> None:
        ...

    def close(self) -> None:
        ...

    def unpause(self) -> None:
        ...

    def moveto(self, n) -> None:
        ...

    def clear(self, nolock: bool = ...) -> None:
        ...

    def refresh(self, nolock: bool = ...) -> None:
        ...

    @property
    def pos(self) -> int:
        ...

    @classmethod
    def set_lock(cls, lock) -> None:
        ...

    @classmethod
    def get_lock(cls) -> None:
        ...

    def set_postfix_dict(self,
                         ordered_dict: None | dict = None,
                         refresh: bool = True,
                         **kwargs) -> None:
        ...

    def set_postfix(self, postfix, **kwargs) -> None:
        ...

    def set_postfix_str(self, s: str = ..., refresh: bool = ...) -> None:
        ...


class _BackwardsCompat:

    @property
    def length(self):
        ...

    @property
    def label(self):
        ...

    def start(self):
        ...

    def stop(self):
        ...


class ProgIter(_TQDMCompat, _BackwardsCompat):
    stream: typing.IO
    iterable: List | Iterable
    desc: str | None
    total: int | None
    freq: int
    initial: int
    enabled: bool
    adjust: bool
    show_percent: bool
    show_times: bool
    show_rate: bool
    show_eta: bool
    show_total: bool
    show_wall: bool
    eta_window: int
    time_thresh: float
    clearline: bool
    chunksize: int | None
    rel_adjust_limit: float
    extra: str
    started: bool
    finished: bool
    homogeneous: bool | str

    def __init__(self,
                 iterable: List | Iterable | None = None,
                 desc: str | None = None,
                 total: int | None = None,
                 freq: int = 1,
                 initial: int = 0,
                 eta_window: int = 64,
                 clearline: bool = True,
                 adjust: bool = True,
                 time_thresh: float = 2.0,
                 show_percent: bool = True,
                 show_times: bool = True,
                 show_rate: bool = True,
                 show_eta: bool = True,
                 show_total: bool = True,
                 show_wall: bool = False,
                 enabled: bool = True,
                 verbose: int | None = None,
                 stream: typing.IO | None = None,
                 chunksize: int | None = None,
                 rel_adjust_limit: float = 4.0,
                 homogeneous: bool | str = 'auto',
                 timer: Callable | None = None,
                 **kwargs) -> None:
        ...

    def __call__(self, iterable: Iterable) -> Iterable:
        ...

    def __enter__(self) -> ProgIter:
        ...

    def __exit__(self, ex_type: Type[BaseException] | None,
                 ex_value: BaseException | None,
                 ex_traceback: TracebackType | None) -> bool | None:
        ...

    def __iter__(self) -> Iterable:
        ...

    def set_extra(self, extra: str | Callable) -> None:
        ...

    def begin(self) -> ProgIter:
        ...

    def end(self) -> None:
        ...

    def step(self, inc: int = 1, force: bool = False) -> None:
        ...

    def format_message(self) -> str:
        ...

    def format_message_parts(self) -> Tuple[str, str, str]:
        ...

    def ensure_newline(self) -> None:
        ...

    def display_message(self) -> None:
        ...



"""
Liberated portions of :mod:`jaraco.windows.filesystem`.

Ignore:

    cat ~/code/ubelt/ubelt/_win32_links.py | grep -o jwfs.api
    cat ~/code/ubelt/ubelt/_win32_links.py | grep -o "jwfs\\.[^ ]*" | sort

    git clone git@github.com:jaraco/jaraco.windows.git $HOME/code
    cd ~/code/jaraco.windows
    touch jaraco/__init__.py

    ---

    Notes:
        liberator does not handle the ctypes attributes nicely where
        the definition is then modified with argtypes and restypes

        But it does help get a good start on the file.

    ---

    import liberator
    import ubelt as ub
    repo_dpath = ub.Path('~/code/jaraco.windows').expand()
    jwfs_modpath = repo_dpath / 'jaraco/windows/filesystem/__init__.py'
    jw_api_filesystem_modpath = repo_dpath / 'jaraco/windows/api/filesystem.py'
    jw_reparse_modpath = repo_dpath / 'jaraco/windows/reparse.py'

    lib = liberator.Liberator()
    lib.add_static('link', modpath=jwfs_modpath)
    lib.add_static('handle_nonzero_success', modpath=jwfs_modpath)
    lib.add_static('is_reparse_point', modpath=jwfs_modpath)

    # FIXME: argtypes / restypes
    lib.add_static('CreateFile', modpath=jw_api_filesystem_modpath)
    lib.add_static('CloseHandle', modpath=jw_api_filesystem_modpath)

    lib.add_static('REPARSE_DATA_BUFFER', modpath=jw_api_filesystem_modpath)
    lib.add_static('OPEN_EXISTING', modpath=jw_api_filesystem_modpath)
    lib.add_static('FILE_FLAG_OPEN_REPARSE_POINT', modpath=jw_api_filesystem_modpath)
    lib.add_static('FILE_FLAG_BACKUP_SEMANTICS', modpath=jw_api_filesystem_modpath)
    lib.add_static('FSCTL_GET_REPARSE_POINT', modpath=jw_api_filesystem_modpath)
    lib.add_static('INVALID_HANDLE_VALUE', modpath=jw_api_filesystem_modpath)
    lib.add_static('IO_REPARSE_TAG_SYMLINK', modpath=jw_api_filesystem_modpath)
    lib.add_static('BY_HANDLE_FILE_INFORMATION', modpath=jw_api_filesystem_modpath)
    lib.add_static('GetFileInformationByHandle', modpath=jw_api_filesystem_modpath)

    #lib.add_static('DeviceIoControl', modpath=jw_reparse_modpath)

    lib.expand(['jaraco'])

    print(lib.current_sourcecode())
"""


import ctypes.wintypes
import ctypes

# Makes mypy happy
import sys
assert sys.platform == "win32"


def handle_nonzero_success(result):
    if (result == 0):
        raise ctypes.WinError()


class BY_HANDLE_FILE_INFORMATION(ctypes.Structure):
    _fields_ = [
        ('file_attributes', ctypes.wintypes.DWORD),
        ('creation_time', ctypes.wintypes.FILETIME),
        ('last_access_time', ctypes.wintypes.FILETIME),
        ('last_write_time', ctypes.wintypes.FILETIME),
        ('volume_serial_number', ctypes.wintypes.DWORD),
        ('file_size_high', ctypes.wintypes.DWORD),
        ('file_size_low', ctypes.wintypes.DWORD),
        ('number_of_links', ctypes.wintypes.DWORD),
        ('file_index_high', ctypes.wintypes.DWORD),
        ('file_index_low', ctypes.wintypes.DWORD)
    ]

    @property
    def file_size(self):
        return ((self.file_size_high << 32) + self.file_size_low)

    @property
    def file_index(self):
        return ((self.file_index_high << 32) + self.file_index_low)


class REPARSE_DATA_BUFFER(ctypes.Structure):
    _fields_ = [
        ('tag', ctypes.c_ulong),
        ('data_length', ctypes.c_ushort),
        ('reserved', ctypes.c_ushort),
        ('substitute_name_offset', ctypes.c_ushort),
        ('substitute_name_length', ctypes.c_ushort),
        ('print_name_offset', ctypes.c_ushort),
        ('print_name_length', ctypes.c_ushort),
        ('flags', ctypes.c_ulong),
        ('path_buffer', (ctypes.c_byte * 1))
    ]

    def get_print_name(self):
        wchar_size = ctypes.sizeof(ctypes.wintypes.WCHAR)
        arr_typ = (ctypes.wintypes.WCHAR * (self.print_name_length // wchar_size))
        data = ctypes.byref(self.path_buffer, self.print_name_offset)
        return ctypes.cast(data, ctypes.POINTER(arr_typ)).contents.value

    def get_substitute_name(self):
        wchar_size = ctypes.sizeof(ctypes.wintypes.WCHAR)
        arr_typ = (ctypes.wintypes.WCHAR * (self.substitute_name_length // wchar_size))
        data = ctypes.byref(self.path_buffer, self.substitute_name_offset)
        return ctypes.cast(data, ctypes.POINTER(arr_typ)).contents.value


class SECURITY_ATTRIBUTES(ctypes.Structure):
    _fields_ = (
        ('length', ctypes.wintypes.DWORD),
        ('p_security_descriptor', ctypes.wintypes.LPVOID),
        ('inherit_handle', ctypes.wintypes.BOOLEAN),
    )

LPSECURITY_ATTRIBUTES = ctypes.POINTER(SECURITY_ATTRIBUTES)


IO_REPARSE_TAG_SYMLINK = 0xA000000C
INVALID_HANDLE_VALUE = ctypes.wintypes.HANDLE((- 1)).value
FSCTL_GET_REPARSE_POINT = 0x900A8
FILE_FLAG_BACKUP_SEMANTICS = 0x2000000
FILE_FLAG_OPEN_REPARSE_POINT = 0x00200000
FILE_SHARE_READ = 1
OPEN_EXISTING = 3


FILE_ATTRIBUTE_REPARSE_POINT = 0x400
GENERIC_READ = 0x80000000
INVALID_FILE_ATTRIBUTES = 0xFFFFFFFF

GetFileAttributes = ctypes.windll.kernel32.GetFileAttributesW
GetFileAttributes.argtypes = (ctypes.wintypes.LPWSTR,)
GetFileAttributes.restype = ctypes.wintypes.DWORD

CreateHardLink = ctypes.windll.kernel32.CreateHardLinkW
CreateHardLink.argtypes = (
    ctypes.wintypes.LPWSTR,
    ctypes.wintypes.LPWSTR,
    ctypes.wintypes.LPVOID,  # reserved for LPSECURITY_ATTRIBUTES
)
CreateHardLink.restype = ctypes.wintypes.BOOLEAN

GetFileInformationByHandle = ctypes.windll.kernel32.GetFileInformationByHandle
GetFileInformationByHandle.restype = ctypes.wintypes.BOOL
GetFileInformationByHandle.argtypes = (
    ctypes.wintypes.HANDLE,
    ctypes.POINTER(BY_HANDLE_FILE_INFORMATION),
)

CloseHandle = ctypes.windll.kernel32.CloseHandle
CloseHandle.argtypes = (ctypes.wintypes.HANDLE,)
CloseHandle.restype = ctypes.wintypes.BOOLEAN

CreateFile = ctypes.windll.kernel32.CreateFileW
CreateFile.argtypes = (
    ctypes.wintypes.LPWSTR,
    ctypes.wintypes.DWORD,
    ctypes.wintypes.DWORD,
    LPSECURITY_ATTRIBUTES,
    ctypes.wintypes.DWORD,
    ctypes.wintypes.DWORD,
    ctypes.wintypes.HANDLE,
)
CreateFile.restype = ctypes.wintypes.HANDLE

LPDWORD = ctypes.POINTER(ctypes.wintypes.DWORD)
LPOVERLAPPED = ctypes.wintypes.LPVOID

DeviceIoControl = ctypes.windll.kernel32.DeviceIoControl
DeviceIoControl.argtypes = [
    ctypes.wintypes.HANDLE,
    ctypes.wintypes.DWORD,
    ctypes.wintypes.LPVOID,
    ctypes.wintypes.DWORD,
    ctypes.wintypes.LPVOID,
    ctypes.wintypes.DWORD,
    LPDWORD,
    LPOVERLAPPED,
]
DeviceIoControl.restype = ctypes.wintypes.BOOL


def is_reparse_point(path):
    """
    Determine if the given path is a reparse point.
    Return False if the file does not exist or the file attributes cannot
    be determined.
    """
    res = GetFileAttributes(path)
    return ((res != INVALID_FILE_ATTRIBUTES) and bool((res & FILE_ATTRIBUTE_REPARSE_POINT)))


def link(target, link):
    """
    Establishes a hard link between an existing file and a new file.
    """
    handle_nonzero_success(CreateHardLink(link, target, None))


def _reparse_DeviceIoControl(device, io_control_code, in_buffer, out_buffer, overlapped=None):
    # ubelt note: name is overloaded, so we mangle it here.
    if overlapped is not None:
        raise NotImplementedError("overlapped handles not yet supported")

    if isinstance(out_buffer, int):
        out_buffer = ctypes.create_string_buffer(out_buffer)

    in_buffer_size = len(in_buffer) if in_buffer is not None else 0
    out_buffer_size = len(out_buffer)
    assert isinstance(out_buffer, ctypes.Array)

    returned_bytes = ctypes.wintypes.DWORD()

    res = DeviceIoControl(
        device,
        io_control_code,
        in_buffer,
        in_buffer_size,
        out_buffer,
        out_buffer_size,
        returned_bytes,
        overlapped,
    )

    handle_nonzero_success(res)
    handle_nonzero_success(returned_bytes)

    return out_buffer[: returned_bytes.value]


# Fake the jaraco api
class api:
    CreateFile                   = CreateFile
    CloseHandle                  = CloseHandle
    GetFileInformationByHandle   = GetFileInformationByHandle

    BY_HANDLE_FILE_INFORMATION   = BY_HANDLE_FILE_INFORMATION
    FILE_FLAG_BACKUP_SEMANTICS   = FILE_FLAG_BACKUP_SEMANTICS
    FILE_FLAG_OPEN_REPARSE_POINT = FILE_FLAG_OPEN_REPARSE_POINT
    FILE_SHARE_READ              = FILE_SHARE_READ
    FSCTL_GET_REPARSE_POINT      = FSCTL_GET_REPARSE_POINT
    GENERIC_READ                 = GENERIC_READ
    INVALID_HANDLE_VALUE         = INVALID_HANDLE_VALUE
    IO_REPARSE_TAG_SYMLINK       = IO_REPARSE_TAG_SYMLINK
    OPEN_EXISTING                = OPEN_EXISTING
    REPARSE_DATA_BUFFER          = REPARSE_DATA_BUFFER


class reparse:
    DeviceIoControl = _reparse_DeviceIoControl



from os.path import exists
import ubelt as ub
import pytest


def test_noexist_meta_clear():
    """
    What no errors happen when an external processes removes meta
    """
    def func():
        return 'expensive result'
    cacher = ub.Cacher('name', 'params', verbose=10)
    cacher.clear()

    cacher.ensure(func)

    data_fpath = cacher.get_fpath()
    meta_fpath = data_fpath + '.meta'
    assert exists(data_fpath)
    assert exists(meta_fpath)

    ub.delete(meta_fpath)
    assert not exists(meta_fpath)
    cacher.clear()

    assert not exists(meta_fpath)
    assert not exists(data_fpath)


def test_clear_quiet():
    """
    What no errors happen when an external processes removes meta
    """
    def func():
        return 'expensive result'
    cacher = ub.Cacher('name', 'params', verbose=0)
    cacher.clear()
    cacher.clear()
    cacher.ensure(func)
    cacher.clear()


def test_corrupt():
    """
    What no errors happen when an external processes removes meta

    python ubelt/tests/test_cache.py test_corrupt
    """
    def func():
        return ['expensive result']
    cacher = ub.Cacher('name', 'params', verbose=10)
    cacher.clear()

    data = cacher.ensure(func)

    data2 = cacher.tryload()

    assert data2 == data

    # Overwrite the data with junk
    with open(cacher.get_fpath(), 'wb') as file:
        file.write(''.encode('utf8'))

    assert cacher.tryload() is None
    with pytest.raises(IOError):
        cacher.load()

    assert cacher.tryload() is None
    with open(cacher.get_fpath(), 'wb') as file:
        file.write(':junkdata:'.encode('utf8'))
    with pytest.raises(Exception):
        cacher.load()


def _setup_corrupt_cacher(verbose=0):
    def func():
        return ['expensive result']
    cacher = ub.Cacher('name', 'params', verbose=verbose)
    cacher.clear()
    cacher.ensure(func)
    # Write junk data that will cause a non-io error
    with open(cacher.get_fpath(), 'wb') as file:
        file.write(':junkdata:'.encode('utf8'))
    with pytest.raises(Exception):
        assert cacher.tryload(on_error='raise')
    assert exists(cacher.get_fpath())
    return cacher


def test_onerror_clear():
    cacher = _setup_corrupt_cacher()
    assert cacher.tryload(on_error='clear') is None
    assert not exists(cacher.get_fpath())
    cacher.clear()


def test_onerror_raise():
    cacher = _setup_corrupt_cacher(verbose=1)
    with pytest.raises(Exception):
        assert cacher.tryload(on_error='raise')
    assert exists(cacher.get_fpath())
    cacher.clear()


def test_onerror_bad_method():
    cacher = _setup_corrupt_cacher()
    assert exists(cacher.get_fpath())
    with pytest.raises(KeyError):
        assert cacher.tryload(on_error='doesnt exist')
    assert exists(cacher.get_fpath())
    cacher.clear()


def test_cache_hit():
    cacher = ub.Cacher('name', 'params', verbose=2)
    cacher.clear()
    assert not cacher.exists()
    cacher.save(['some', 'data'])
    assert cacher.exists()
    data = cacher.load()
    assert data == ['some', 'data']


def test_disable():
    """
    What no errors happen when an external processes removes meta
    """
    nonlocal_var = [0]

    def func():
        nonlocal_var[0] += 1
        return ['expensive result']
    cacher = ub.Cacher('name', 'params', verbose=10, enabled=False)

    assert nonlocal_var[0] == 0
    cacher.ensure(func)
    assert nonlocal_var[0] == 1
    cacher.ensure(func)
    assert nonlocal_var[0] == 2
    cacher.ensure(func)

    with pytest.raises(IOError):
        cacher.load()

    assert cacher.tryload(func) is None


def test_disabled_cache_stamp():
    stamp = ub.CacheStamp('foo', 'bar', enabled=False)
    assert stamp.expired() == 'disabled', 'disabled cache stamps are always expired'


def test_cache_depends():
    """
    What no errors happen when an external processes removes meta
    """
    cacher = ub.Cacher('name', depends=['a', 'b', 'c'],
                        verbose=10, enabled=False)
    cfgstr = cacher._rectify_cfgstr()
    assert cfgstr.startswith('8a82eef87cb905220841f95')


def test_cache_cfgstr():
    """
    TODO: remove when cfgstr is removed
    """
    # with pytest.warns(DeprecationWarning):
    with pytest.raises(RuntimeError):
        cacher1 = ub.Cacher('name', cfgstr='abc')
        cacher1
        # assert cacher1.depends == 'abc'


def test_cache_stamp_with_hash():
    dpath = ub.Path.appdir('ubelt/tests/test-cache-stamp-with-hash')
    for verbose in [0, 1]:
        dpath.delete().ensuredir()
        fpath = dpath / 'result.txt'
        fpath.write_text('hello')
        expected_hash = ub.hash_file(fpath, hasher='sha256')
        unexpected_hash = 'fdsfdsafds'
        stamp = ub.CacheStamp(
            'foo.stamp', dpath=dpath, product=[fpath], depends='nodep',
            hash_prefix=unexpected_hash, verbose=verbose, hasher='sha256', ext='.json')
        assert not exists(stamp.cacher.get_fpath())
        with pytest.raises(RuntimeError):
            stamp.renew()
        assert not exists(stamp.cacher.get_fpath())
        # Fix the expected hash and now renew should work
        stamp.hash_prefix = expected_hash
        stamp.renew()
        assert exists(stamp.cacher.get_fpath())
        assert not stamp.expired()

        # But change it back, and we will be expired
        stamp.hash_prefix = unexpected_hash
        assert stamp.expired() == 'hash_prefix_mismatch'

        # Then change it back, and we are ok
        stamp.hash_prefix = expected_hash
        assert not stamp.expired()

        # Corrupt the file and check for hash diff
        # (need to disable mtime check for this to always work)
        stamp._expire_checks['mtime'] = False
        fpath.write_text('jello')
        assert stamp.expired() == 'hash_diff'

        # Disabling the hash check makes us rely on size / mtime, but is faster
        stamp._expire_checks['hash'] = False
        assert not stamp.expired()

if __name__ == '__main__':
    r"""
    CommandLine:
        pytest ubelt/tests/test_cache.py
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



"""
TODO: test _can_symlink=False variants on systems that can symlink.
"""
from os.path import isdir
from os.path import isfile
from os.path import islink
from os.path import join, exists, relpath, dirname
import ubelt as ub
import pytest
import os
from ubelt import util_links
import sys


if sys.platform.startswith('win32'):
    try:
        import jaraco.windows.filesystem as jwfs
    except ImportError:
        jwfs = None


def test_rel_dir_link():
    """
    xdoctest ~/code/ubelt/tests/test_links.py test_rel_dir_link
    """
    import pytest
    import ubelt as ub
    if ub.WIN32 and jwfs is None:
        pytest.skip()  # hack for windows for now.

    dpath = ub.Path.appdir('ubelt/tests/test_links', 'test_rel_dir_link').ensuredir()
    ub.delete(dpath, verbose=2)
    ub.ensuredir(dpath, verbose=2)

    real_dpath = join((dpath / 'dir1').ensuredir(), 'real')
    link_dpath = join((dpath / 'dir2').ensuredir(), 'link')
    ub.ensuredir(real_dpath)

    orig = os.getcwd()
    try:
        os.chdir(dpath)
        real_path = relpath(real_dpath, dpath)
        link_path = relpath(link_dpath, dpath)
        link = ub.symlink(real_path, link_path)
        # Note: on windows this is hacked.
        pointed = ub.util_links._readlink(link)
        resolved = os.path.realpath(ub.expandpath(join(dirname(link), pointed)))

        final_real_dpath = os.path.realpath(ub.expandpath(real_dpath))
        if final_real_dpath != resolved:
            raise AssertionError(f'{final_real_dpath} != {resolved}')
        # assert os.path.realpath(ub.expandpath(real_dpath)) == resolved
    except Exception:
        util_links._dirstats(dpath)
        util_links._dirstats(join(dpath, 'dir1'))
        util_links._dirstats(join(dpath, 'dir2'))
        print('TEST FAILED: test_rel_link')
        print('real_dpath = {!r}'.format(real_dpath))
        print('link_dpath = {!r}'.format(link_dpath))
        print('real_path = {!r}'.format(real_path))
        print('link_path = {!r}'.format(link_path))
        try:
            if 'link' in vars():
                print('link = {!r}'.format(link))
            if 'pointed' in vars():
                print('pointed = {!r}'.format(pointed))
            if 'resolved' in vars():
                print('resolved = {!r}'.format(resolved))
        except Exception:
            print('...rest of the names are not available')
        raise
    finally:
        util_links._dirstats(dpath)
        util_links._dirstats(join(dpath, 'dir1'))
        util_links._dirstats(join(dpath, 'dir2'))
        os.chdir(orig)


def test_rel_file_link():
    import pytest
    import ubelt as ub
    if ub.WIN32 and jwfs is None:
        pytest.skip()  # hack for windows for now.
    dpath = ub.Path.appdir('ubelt/tests/test_links', 'test_rel_file_link').ensuredir()
    ub.delete(dpath, verbose=2)
    ub.ensuredir(dpath, verbose=2)

    real_fpath = join(ub.ensuredir((dpath, 'dir1')), 'real')
    link_fpath = join(ub.ensuredir((dpath, 'dir2')), 'link')
    ub.touch(real_fpath)

    orig = os.getcwd()
    try:
        os.chdir(dpath)
        real_path = relpath(real_fpath, dpath)
        link_path = relpath(link_fpath, dpath)
        link = ub.symlink(real_path, link_path)
        import sys
        if sys.platform.startswith('win32') and isfile(link):
            # Note: if windows hard links the file there is no way we can
            # tell that it was a symlink. Just verify it exists.
            from ubelt import _win32_links
            assert _win32_links._win32_is_hardlinked(real_fpath, link_fpath)
        else:
            pointed = ub.util_links._readlink(link)
            resolved = os.path.realpath(ub.expandpath(join(dirname(link), pointed)))
            assert os.path.realpath(ub.expandpath(real_fpath)) == resolved
    except Exception:
        util_links._dirstats(dpath)
        util_links._dirstats(join(dpath, 'dir1'))
        util_links._dirstats(join(dpath, 'dir2'))
        print('TEST FAILED: test_rel_link')
        print('real_fpath = {!r}'.format(real_fpath))
        print('link_fpath = {!r}'.format(link_fpath))
        print('real_path = {!r}'.format(real_path))
        print('link_path = {!r}'.format(link_path))
        try:
            if 'link' in vars():
                print('link = {!r}'.format(link))
            if 'pointed' in vars():
                print('pointed = {!r}'.format(pointed))
            if 'resolved' in vars():
                print('resolved = {!r}'.format(resolved))
        except Exception:
            print('...rest of the names are not available')
        raise
    finally:
        util_links._dirstats(dpath)
        util_links._dirstats(join(dpath, 'dir1'))
        util_links._dirstats(join(dpath, 'dir2'))
        os.chdir(orig)


def test_delete_symlinks():
    """
    CommandLine:
        python -m ubelt.tests.test_links test_delete_symlinks
    """
    import pytest
    import ubelt as ub
    if ub.WIN32 and jwfs is None:
        pytest.skip()  # hack for windows for now.
    # TODO: test that we handle broken links
    dpath = ub.Path.appdir('ubelt/tests/test_links', 'test_delete_links').ensuredir()

    happy_dpath = join(dpath, 'happy_dpath')
    happy_dlink = join(dpath, 'happy_dlink')
    happy_fpath = join(dpath, 'happy_fpath.txt')
    happy_flink = join(dpath, 'happy_flink.txt')
    broken_dpath = join(dpath, 'broken_dpath')
    broken_dlink = join(dpath, 'broken_dlink')
    broken_fpath = join(dpath, 'broken_fpath.txt')
    broken_flink = join(dpath, 'broken_flink.txt')

    def check_path_condition(path, positive, want, msg):
        if not want:
            positive = not positive
            msg = 'not ' + msg

        if not positive:
            util_links._dirstats(dpath)
            print('About to raise error: {}'.format(msg))
            print('path = {!r}'.format(path))
            print('exists(path) = {!r}'.format(exists(path)))
            print('islink(path) = {!r}'.format(islink(path)))
            print('isdir(path) = {!r}'.format(isdir(path)))
            print('isfile(path) = {!r}'.format(isfile(path)))
            raise AssertionError('path={} {}'.format(path, msg))

    def assert_sometrace(path, want=True):
        # Either exists or is a broken link
        positive = exists(path) or islink(path)
        check_path_condition(path, positive, want, 'has trace')

    def assert_broken_link(path, want=True):
        if util_links._can_symlink():
            print('path={} should{} be a broken link'.format(
                path, ' ' if want else ' not'))
            positive = not exists(path) and islink(path)
            check_path_condition(path, positive, want, 'broken link')
        else:
            # TODO: we can test this
            # positive = util_links._win32_is_junction(path)
            print('path={} should{} be a broken link (junction)'.format(
                path, ' ' if want else ' not'))
            print('cannot check this yet')
            # We wont be able to differentiate links and nonlinks for junctions
            # positive = exists(path)
            # check_path_condition(path, positive, want, 'broken link')

    util_links._dirstats(dpath)
    ub.delete(dpath, verbose=2)
    ub.ensuredir(dpath, verbose=2)
    util_links._dirstats(dpath)

    ub.ensuredir(happy_dpath, verbose=2)
    ub.ensuredir(broken_dpath, verbose=2)
    ub.touch(happy_fpath, verbose=2)
    ub.touch(broken_fpath, verbose=2)
    util_links._dirstats(dpath)

    ub.symlink(broken_fpath, broken_flink, verbose=2)
    ub.symlink(broken_dpath, broken_dlink, verbose=2)
    ub.symlink(happy_fpath, happy_flink, verbose=2)
    ub.symlink(happy_dpath, happy_dlink, verbose=2)
    util_links._dirstats(dpath)

    # Deleting the files should not delete the symlinks (windows)
    ub.delete(broken_fpath, verbose=2)
    util_links._dirstats(dpath)

    ub.delete(broken_dpath, verbose=2)
    util_links._dirstats(dpath)

    assert_broken_link(broken_flink, 1)
    assert_broken_link(broken_dlink, 1)
    assert_sometrace(broken_fpath, 0)
    assert_sometrace(broken_dpath, 0)

    assert_broken_link(happy_flink, 0)
    assert_broken_link(happy_dlink, 0)
    assert_sometrace(happy_fpath, 1)
    assert_sometrace(happy_dpath, 1)

    # broken symlinks no longer exist after they are deleted
    ub.delete(broken_dlink, verbose=2)
    util_links._dirstats(dpath)
    assert_sometrace(broken_dlink, 0)

    ub.delete(broken_flink, verbose=2)
    util_links._dirstats(dpath)
    assert_sometrace(broken_flink, 0)

    # real symlinks no longer exist after they are deleted
    # but the original data is fine
    ub.delete(happy_dlink, verbose=2)
    util_links._dirstats(dpath)
    assert_sometrace(happy_dlink, 0)
    assert_sometrace(happy_dpath, 1)

    ub.delete(happy_flink, verbose=2)
    util_links._dirstats(dpath)
    assert_sometrace(happy_flink, 0)
    assert_sometrace(happy_fpath, 1)


def test_modify_directory_symlinks():
    import pytest
    import ubelt as ub
    if ub.WIN32 and jwfs is None:
        pytest.skip()  # hack for windows for now.
    dpath = ub.Path.appdir('ubelt/tests/test_links', 'test_modify_symlinks').ensuredir()
    ub.delete(dpath, verbose=2)
    ub.ensuredir(dpath, verbose=2)

    happy_dpath = dpath / 'happy_dpath'
    happy_dlink = dpath / 'happy_dlink'
    ub.ensuredir(happy_dpath, verbose=2)

    ub.symlink(happy_dpath, happy_dlink, verbose=2)

    # Test file inside directory symlink
    file_path1 = happy_dpath / 'file.txt'
    file_path2 = happy_dlink / 'file.txt'

    ub.touch(file_path1, verbose=2)
    assert file_path1.exists()
    assert file_path2.exists()

    file_path1.write_text('foo')
    assert file_path1.read_text() == 'foo'
    assert file_path2.read_text() == 'foo'

    file_path2.write_text('bar')
    assert file_path1.read_text() == 'bar'
    assert file_path2.read_text() == 'bar'

    ub.delete(file_path2, verbose=2)
    assert not file_path1.exists()
    assert not file_path2.exists()

    # Test directory inside directory symlink
    dir_path1 = happy_dpath / 'dir'
    dir_path2 = happy_dlink / 'dir'

    ub.ensuredir(dir_path1, verbose=2)
    assert dir_path1.exists()
    assert dir_path2.exists()

    subfile_path1 = dir_path1 / 'subfile.txt'
    subfile_path2 = dir_path2 / 'subfile.txt'

    subfile_path1.write_text('foo')
    assert subfile_path1.read_text() == 'foo'
    assert subfile_path2.read_text() == 'foo'

    subfile_path1.write_text('bar')
    assert subfile_path1.read_text() == 'bar'
    assert subfile_path2.read_text() == 'bar'

    ub.delete(dir_path1, verbose=2)
    assert not dir_path1.exists()
    assert not dir_path2.exists()


def test_modify_file_symlinks():
    """
    CommandLine:
        python -m ubelt.tests.test_links test_modify_symlinks
    """
    import pytest
    import ubelt as ub
    if ub.WIN32 and jwfs is None:
        pytest.skip()  # hack for windows for now.
    # TODO: test that we handle broken links
    dpath = ub.Path.appdir('ubelt/tests/test_links', 'test_modify_symlinks').ensuredir()
    happy_fpath = dpath / 'happy_fpath.txt'
    happy_flink = dpath / 'happy_flink.txt'
    ub.touch(happy_fpath, verbose=2)

    ub.symlink(happy_fpath, happy_flink, verbose=2)

    # Test file symlink
    happy_fpath.write_text('foo')
    assert happy_fpath.read_text() == 'foo'
    assert happy_flink.read_text() == 'foo'

    happy_flink.write_text('bar')
    assert happy_fpath.read_text() == 'bar'
    assert happy_flink.read_text() == 'bar'


def test_broken_link():
    """
    CommandLine:
        python -m ubelt.tests.test_links test_broken_link
    """
    import pytest
    import ubelt as ub
    if ub.WIN32 and jwfs is None:
        pytest.skip()  # hack for windows for now.
    dpath = ub.Path.appdir('ubelt/tests/test_links', 'test_broken_link').ensuredir()

    ub.delete(dpath, verbose=2)
    ub.ensuredir(dpath, verbose=2)
    util_links._dirstats(dpath)

    broken_fpath = join(dpath, 'broken_fpath.txt')
    broken_flink = join(dpath, 'broken_flink.txt')

    ub.touch(broken_fpath, verbose=2)
    util_links._dirstats(dpath)
    ub.symlink(real_path=broken_fpath, link_path=broken_flink, verbose=2)

    util_links._dirstats(dpath)
    ub.delete(broken_fpath, verbose=2)

    util_links._dirstats(dpath)

    # make sure I am sane that this is the correct check.
    can_symlink = util_links._can_symlink()
    print('can_symlink = {!r}'.format(can_symlink))
    if can_symlink:
        # normal behavior
        assert islink(broken_flink)
        assert not exists(broken_flink)
    else:
        # on windows hard links are essentially the same file.
        # there is no trace that it was actually a link.
        assert exists(broken_flink)


def test_cant_overwrite_file_with_symlink():
    if ub.WIN32:
        # Can't distinguish this case on windows
        pytest.skip()

    dpath = ub.Path.appdir('ubelt/tests/test_links', 'test_cant_overwrite_file_with_symlink').ensuredir()
    ub.delete(dpath, verbose=2)
    ub.ensuredir(dpath, verbose=2)

    happy_fpath = join(dpath, 'happy_fpath.txt')
    happy_flink = join(dpath, 'happy_flink.txt')

    for verbose in [2, 1, 0]:
        print('=======')
        print('verbose = {!r}'.format(verbose))
        ub.delete(dpath, verbose=verbose)
        ub.ensuredir(dpath, verbose=verbose)
        ub.touch(happy_fpath, verbose=verbose)
        ub.touch(happy_flink)  # create a file where a link should be

        util_links._dirstats(dpath)
        with pytest.raises(FileExistsError):  # file exists error
            ub.symlink(happy_fpath, happy_flink, overwrite=False, verbose=verbose)

        with pytest.raises(FileExistsError):  # file exists error
            ub.symlink(happy_fpath, happy_flink, overwrite=True, verbose=verbose)


def test_overwrite_symlink():
    """
    CommandLine:
        python ~/code/ubelt/tests/test_links.py test_overwrite_symlink
    """
    import pytest
    import ubelt as ub
    if ub.WIN32 and jwfs is None:
        pytest.skip()  # hack for windows for now.

    # TODO: test that we handle broken links
    dpath = ub.Path.appdir('ubelt/tests/test_links', 'test_overwrite_symlink').ensuredir()
    ub.delete(dpath, verbose=2)
    ub.ensuredir(dpath, verbose=2)

    happy_fpath = join(dpath, 'happy_fpath.txt')
    other_fpath = join(dpath, 'other_fpath.txt')
    happy_flink = join(dpath, 'happy_flink.txt')

    for verbose in [2, 1, 0]:
        print('@==========@')
        print('verbose = {!r}'.format(verbose))

        print('[test] Setup')
        ub.delete(dpath, verbose=verbose)
        ub.ensuredir(dpath, verbose=verbose)
        ub.touch(happy_fpath, verbose=verbose)
        ub.touch(other_fpath, verbose=verbose)

        print('[test] Dirstats dpath')
        util_links._dirstats(dpath)

        print('[test] Create initial link (to happy)')
        ub.symlink(happy_fpath, happy_flink, verbose=verbose)

        print('[test] Dirstats dpath')
        util_links._dirstats(dpath)

        # Creating a duplicate link
        print('[test] Create a duplicate link (to happy)')
        ub.symlink(happy_fpath, happy_flink, verbose=verbose)

        print('[test] Dirstats dpath')
        util_links._dirstats(dpath)

        print('[test] Create an unauthorized overwrite link (to other)')
        with pytest.raises(Exception) as exc_info:  # file exists error
            ub.symlink(other_fpath, happy_flink, verbose=verbose)
        print(' * exc_info = {!r}'.format(exc_info))

        print('[test] Create an authorized overwrite link (to other)')
        ub.symlink(other_fpath, happy_flink, verbose=verbose, overwrite=True)

        print('[test] Dirstats dpath')
        ub.delete(other_fpath, verbose=verbose)

        print('[test] Create an unauthorized overwrite link (back to happy)')
        with pytest.raises(Exception) as exc_info:  # file exists error
            ub.symlink(happy_fpath, happy_flink, verbose=verbose)
        print(' * exc_info = {!r}'.format(exc_info))

        print('[test] Create an authorized overwrite link (back to happy)')
        ub.symlink(happy_fpath, happy_flink, verbose=verbose, overwrite=True)


def _force_junction(func):
    from functools import wraps
    @wraps(func)
    def _wrap(*args):
        if not ub.WIN32:
            pytest.skip()
        from ubelt import _win32_links
        _win32_links.__win32_can_symlink__ = False
        func(*args)
        _win32_links.__win32_can_symlink__ = None
    return _wrap


def test_symlink_to_rel_symlink():
    """
    Test a case with a absolute link to a relative link to a real path.
    """
    import ubelt as ub
    if ub.WIN32:
        import pytest
        pytest.skip('dont try on windows')

    dpath = ub.Path.appdir('ubelt/tests/links/sym-to-relsym')
    dpath.delete().ensuredir()

    level1 = (dpath / 'level1').ensuredir()

    real = dpath / 'real'
    link1 = level1 / 'link1'

    real.touch()

    print('Should create')

    # rel_link1_to_real = os.path.relpath(real, link1.parent)
    # FIXME: This ub.symlink behavior seems broken

    link1.symlink_to(os.path.relpath(real, link1.parent))
    # ub.symlink(real_path=rel_link1_to_real, link_path=link1, verbose=3)

    # ub.symlink(real_path=rel_link1_to_real, link_path=link2, verbose=3)
    """
    At this point we have:

    ├── level1
    │   ├── level2
    │   │   └── link2 -> /home/joncrall/.cache/ubelt/tests/links/sym-to-relsym/level1/link1
    │   └── link1 -> ../real
    └── real
    """
    # _ = ub.cmd(f'tree {dpath}', verbose=3)

    import pytest
    with pytest.raises(FileExistsError):
        ub.symlink(real_path=real, link_path=link1, verbose=3)

    # ub.symlink(real_path=link1, link_path=link2, verbose=1)


# class TestSymlinksForceJunction(object):
fj_test_delete_symlinks = _force_junction(test_delete_symlinks)
fj_test_modify_directory_symlinks = _force_junction(test_modify_directory_symlinks)
fj_test_modify_file_symlinks = _force_junction(test_modify_file_symlinks)
fj_test_broken_link = _force_junction(test_broken_link)
fj_test_overwrite_symlink = _force_junction(test_overwrite_symlink)


if __name__ == '__main__':
    r"""
    CommandLine:
        set PYTHONPATH=%PYTHONPATH%;C:/Users/erote/code/ubelt/ubelt/tests
        pytest ubelt/tests/test_links.py
        pytest ubelt/tests/test_links.py -s
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



from typing import Tuple


class NoParamType:

    def __new__(cls) -> NoParamType:
        ...

    def __reduce__(self) -> Tuple[type, Tuple]:
        ...

    def __copy__(self) -> NoParamType:
        ...

    def __deepcopy__(self, memo) -> NoParamType:
        ...

    def __bool__(self) -> bool:
        ...


NoParam: NoParamType



"""
Functions for capturing and redirecting IO streams with optional
tee-functionality.

The :class:`CaptureStdout` captures all text sent to stdout and optionally
prevents it from actually reaching stdout.

The :class:`TeeStringIO` does the same thing but for arbitrary streams. It is
how the former is implemented.

"""
import sys
import io

__all__ = [
    'TeeStringIO',
    'CaptureStdout',
    'CaptureStream',
]


class TeeStringIO(io.StringIO):
    """
    An IO object that writes to itself and another IO stream.

    Attributes:
        redirect (io.IOBase | None): The other stream to write to.

    Example:
        >>> import ubelt as ub
        >>> import io
        >>> redirect = io.StringIO()
        >>> self = ub.TeeStringIO(redirect)
        >>> self.write('spam')
        >>> assert self.getvalue() == 'spam'
        >>> assert redirect.getvalue() == 'spam'
    """
    def __init__(self, redirect=None):
        """
        Args:
            redirect (io.IOBase): The other stream to write to.
        """
        self.redirect = redirect  # type: io.IOBase
        super().__init__()

        # Logic taken from prompt_toolkit/output/vt100.py version 3.0.5 in
        # flush I don't have a full understanding of what the buffer
        # attribute is supposed to be capturing here, but this seems to
        # allow us to embed in IPython while still capturing and Teeing
        # stdout
        if hasattr(redirect, 'buffer'):
            self.buffer = redirect.buffer  # Py3.
        else:
            self.buffer = redirect
        # Note: mypy doesn't like this type
        # buffer (io.BufferedIOBase | io.IOBase | None): the redirected buffer attribute

    def isatty(self):  # nocover
        """
        Returns true of the redirect is a terminal.

        Note:
            Needed for ``IPython.embed`` to work properly when this class is
            used to override stdout / stderr.

        SeeAlso:
            :meth:`io.IOBase.isatty`

        Returns:
            bool
        """
        return (self.redirect is not None and
                hasattr(self.redirect, 'isatty') and self.redirect.isatty())

    def fileno(self):
        """
        Returns underlying file descriptor of the redirected IOBase object
        if one exists.

        Returns:
            int : the integer corresponding to the file descriptor

        SeeAlso:
            :meth:`io.IOBase.fileno`

        Example:
            >>> import ubelt as ub
            >>> dpath = ub.Path.appdir('ubelt/tests/util_stream').ensuredir()
            >>> fpath = dpath / 'fileno-test.txt'
            >>> with open(fpath, 'w') as file:
            >>>     self = ub.TeeStringIO(file)
            >>>     descriptor = self.fileno()
            >>>     print(f'descriptor={descriptor}')
            >>>     assert isinstance(descriptor, int)

        Example:
            >>> # Test errors
            >>> # Not sure the best way to test, this func is important for
            >>> # capturing stdout when ipython embedding
            >>> import io
            >>> import pytest
            >>> import ubelt as ub
            >>> with pytest.raises(io.UnsupportedOperation):
            >>>     ub.TeeStringIO(redirect=io.StringIO()).fileno()
            >>> with pytest.raises(io.UnsupportedOperation):
            >>>     ub.TeeStringIO(None).fileno()
        """
        if self.redirect is not None:
            return self.redirect.fileno()
        else:
            return super().fileno()

    @property
    def encoding(self):
        """
        Gets the encoding of the `redirect` IO object

        FIXME:
            My complains that this violates the Liskov substitution principle
            because the return type can be str or None, whereas the parent
            class always returns a None. In the future we may raise an exception
            instead of returning None.

        SeeAlso:
            :py:obj:`io.TextIOBase.encoding`

        Example:
            >>> import ubelt as ub
            >>> redirect = io.StringIO()
            >>> assert ub.TeeStringIO(redirect).encoding is None
            >>> assert ub.TeeStringIO(None).encoding is None
            >>> assert ub.TeeStringIO(sys.stdout).encoding is sys.stdout.encoding
            >>> redirect = io.TextIOWrapper(io.StringIO())
            >>> assert ub.TeeStringIO(redirect).encoding is redirect.encoding
        """
        # mypy correctly complains if we include the return type, but we need
        # to keep this buggy behavior for legacy reasons.
        # Returns:
        #     None | str
        if self.redirect is not None:
            return self.redirect.encoding
        else:
            return super().encoding

    @encoding.setter
    def encoding(self, value):
        # Adding a setter to make mypy happy
        raise Exception('Cannot set encoding attribute')

    def write(self, msg):
        """
        Write to this and the redirected stream

        Args:
            msg (str): the data to write

        SeeAlso:
            :meth:`io.TextIOBase.write`

        Example:
            >>> import ubelt as ub
            >>> dpath = ub.Path.appdir('ubelt/tests/util_stream').ensuredir()
            >>> fpath = dpath / 'write-test.txt'
            >>> with open(fpath, 'w') as file:
            >>>     self = ub.TeeStringIO(file)
            >>>     n = self.write('hello world')
            >>>     assert n == 11
            >>> assert self.getvalue() == 'hello world'
            >>> assert fpath.read_text() == 'hello world'
        """
        if self.redirect is not None:
            self.redirect.write(msg)
        return super().write(msg)

    def flush(self):  # nocover
        """
        Flush to this and the redirected stream

        SeeAlso:
            :meth:`io.IOBase.flush`
        """
        if self.redirect is not None:
            self.redirect.flush()
        return super().flush()


class CaptureStream(object):
    """
    Generic class for capturing streaming output from stdout or stderr
    """


class CaptureStdout(CaptureStream):
    r"""
    Context manager that captures stdout and stores it in an internal stream.

    Depending on the value of ``suppress``, the user can control if stdout is
    printed (i.e. if stdout is tee-ed or suppressed) while it is being captured.

    SeeAlso:
        :func:`contextlib.redirect_stdout` - similar, but does not have the
            ability to print stdout while it is being captured.

    Attributes:
        text (str | None): internal storage for the most recent part

        parts (List[str]): internal storage for all parts

        cap_stdout (None | TeeStringIO): internal stream proxy

        orig_stdout (io.TextIOBase):
            internal pointer to the original stdout stream

    Example:
        >>> import ubelt as ub
        >>> self = ub.CaptureStdout(suppress=True)
        >>> print('dont capture the table flip (╯°□°）╯︵ ┻━┻')
        >>> with self:
        ...     text = 'capture the heart ♥'
        ...     print(text)
        >>> print('dont capture look of disapproval ಠ_ಠ')
        >>> assert isinstance(self.text, str)
        >>> assert self.text == text + '\n', 'failed capture text'

    Example:
        >>> import ubelt as ub
        >>> self = ub.CaptureStdout(suppress=False)
        >>> with self:
        ...     print('I am captured and printed in stdout')
        >>> assert self.text.strip() == 'I am captured and printed in stdout'

    Example:
        >>> import ubelt as ub
        >>> self = ub.CaptureStdout(suppress=True, enabled=False)
        >>> with self:
        ...     print('dont capture')
        >>> assert self.text is None
    """
    def __init__(self, suppress=True, enabled=True):
        """
        Args:
            suppress (bool):
                if True, stdout is not printed while captured.
                Defaults to True.

            enabled (bool):
                does nothing if this is False. Defaults to True.
        """
        self.text = None
        self._pos = 0  # keep track of how much has been logged
        self.parts = []
        self.started = False
        self.cap_stdout = None
        self.enabled = enabled
        self.suppress = suppress
        self.orig_stdout = sys.stdout

        if suppress:
            redirect = None
        else:
            redirect = self.orig_stdout
        self.cap_stdout = TeeStringIO(redirect)

    def log_part(self):
        """ Log what has been captured so far """
        self.cap_stdout.seek(self._pos)
        text = self.cap_stdout.read()
        self._pos = self.cap_stdout.tell()
        self.parts.append(text)
        self.text = text

    def start(self):
        if self.enabled:
            self.text = ''
            self.started = True
            sys.stdout = self.cap_stdout

    def stop(self):
        """
        Example:
            >>> import ubelt as ub
            >>> ub.CaptureStdout(enabled=False).stop()
            >>> ub.CaptureStdout(enabled=True).stop()
        """
        if self.enabled:
            self.started = False
            sys.stdout = self.orig_stdout

    def __enter__(self):
        self.start()
        return self

    def __del__(self):  # nocover
        if self.started:
            self.stop()
        if self.cap_stdout is not None:
            self.close()

    def close(self):
        self.cap_stdout.close()
        self.cap_stdout = None

    def __exit__(self, ex_type, ex_value, ex_traceback):
        """
        Args:
            ex_type (Type[BaseException] | None):
            ex_value (BaseException | None):
            ex_traceback (TracebackType | None):

        Returns:
            bool | None
        """
        if self.enabled:
            try:
                self.log_part()
            finally:
                self.stop()
        if ex_traceback is not None:
            return False  # return a falsey value on error



def test_job_pool_context_manager():
    import ubelt as ub

    def worker(data):
        return data + 1

    pool = ub.JobPool('thread', max_workers=16)
    with pool:

        for data in ub.ProgIter(range(10), desc='submit jobs'):
            pool.submit(worker, data)

        final = []
        for job in pool.as_completed(desc='collect jobs'):
            info = job.result()
            final.append(info)


def test_job_pool_as_completed_prog_args():
    import ubelt as ub

    def worker(data):
        return data + 1

    pool = ub.JobPool('thread', max_workers=1)

    for data in ub.ProgIter(range(10), desc='submit jobs'):
        pool.submit(worker, data)

    with ub.CaptureStdout() as cap:
        final = list(pool.as_completed(desc='collect jobs', progkw={'verbose': 3, 'time_thresh': 0}))

    print(f'cap.text={cap.text}')
    num_lines = len(cap.text.split('\n'))
    num_jobs = len(pool.jobs)
    assert num_lines > num_jobs

    print('final = {!r}'.format(final))
    pool.shutdown()


def test_executor_timeout():
    import pytest
    pytest.skip(
        'long test, demos that timeout does not work with SerialExecutor')

    import ubelt as ub
    import time
    from concurrent.futures import TimeoutError

    def long_job(n, t):
        for i in range(n):
            time.sleep(t)

    for mode in ['thread', 'process', 'serial']:
        executor = ub.Executor(mode=mode, max_workers=1)
        with executor:
            job = executor.submit(long_job, 10, 0.05)
            with ub.Timer() as timer:
                try:
                    job_result = job.result(timeout=0.01)
                except TimeoutError as ex:
                    ex_ = ex
                else:
                    print('job_result = {!r}'.format(job_result))
            print('timer.elapsed = {!r}'.format(timer.elapsed))
            print('ex_ = {!r}'.format(ex_))


def test_job_pool_clear_completed():
    import weakref
    import gc
    import ubelt as ub
    is_deleted = {}
    weak_futures = {}

    jobs = ub.JobPool(mode='process', max_workers=4)

    def make_finalizer(jobid):
        def _finalizer():
            is_deleted[jobid] = True
        return _finalizer

    def debug_referrers():
        if 0:
            referrers = ub.udict({})
            for jobid, ref in weak_futures.items():
                fs = ref()
                referrers[jobid] = 0 if fs is None else len(gc.get_referrers(fs))
            print('is_deleted = {}'.format(ub.urepr(is_deleted, nl=1)))
            print('referrers = {}'.format(ub.urepr(referrers, nl=1)))

    for jobid in range(10):
        fs = jobs.submit(simple_worker, jobid)
        weak_futures[jobid] = weakref.ref(fs)
        is_deleted[jobid] = False
        weakref.finalize(fs, make_finalizer(jobid))
        del fs

    debug_referrers()
    assert not any(is_deleted.values())

    for fs in jobs.as_completed():
        fs.result()

    debug_referrers()
    assert not any(is_deleted.values())

    jobs._clear_completed()

    debug_referrers()

    import platform
    if 'pypy' not in platform.python_implementation().lower():
        if not any(is_deleted.values()):
            raise AssertionError

    fs = None

    if 'pypy' not in platform.python_implementation().lower():
        if not all(is_deleted.values()):
            raise AssertionError


def simple_worker(jobid):
    return jobid


def test_job_pool_transient():
    import weakref
    import ubelt as ub
    is_deleted = {}
    weak_futures = {}

    jobs = ub.JobPool(mode='process', max_workers=4, transient=True)

    def make_finalizer(jobid):
        def _finalizer():
            is_deleted[jobid] = True
        return _finalizer

    for jobid in range(10):
        fs = jobs.submit(simple_worker, jobid)
        weak_futures[jobid] = weakref.ref(fs)
        is_deleted[jobid] = False
        weakref.finalize(fs, make_finalizer(jobid))

    if any(is_deleted.values()):
        raise AssertionError

    for fs in jobs.as_completed():
        fs.result()

    # For 3.6, pytest has an AST issue if and assert statements are used.
    # raising regular AssertionErrors to handle that.
    import platform
    if 'pypy' not in platform.python_implementation().lower():
        if not any(is_deleted.values()):
            raise AssertionError

    fs = None

    if 'pypy' not in platform.python_implementation().lower():
        if not all(is_deleted.values()):
            raise AssertionError


def test_backends():
    import platform
    import sys
    # The process backend breaks pyp3 when using coverage
    if 'pypy' in platform.python_implementation().lower():
        import pytest
        pytest.skip('not testing process on pypy')
    if sys.platform.startswith('win32'):
        import pytest
        pytest.skip('not running this test on win32 for now')
    import ubelt as ub
    # Fork before threading!
    # https://pybay.com/site_media/slides/raymond2017-keynote/combo.html
    self1 = ub.Executor(mode='serial', max_workers=0)
    self1.__enter__()
    self2 = ub.Executor(mode='process', max_workers=2)
    self2.__enter__()
    self3 = ub.Executor(mode='thread', max_workers=2)
    self3.__enter__()
    jobs = []
    jobs.append(self1.submit(sum, [1, 2, 3]))
    jobs.append(self1.submit(sum, [1, 2, 3]))
    jobs.append(self2.submit(sum, [10, 20, 30]))
    jobs.append(self2.submit(sum, [10, 20, 30]))
    jobs.append(self3.submit(sum, [4, 5, 5]))
    jobs.append(self3.submit(sum, [4, 5, 5]))
    for job in jobs:
        result = job.result()
        print('result = {!r}'.format(result))
    self1.__exit__(None, None, None)
    self2.__exit__(None, None, None)
    self3.__exit__(None, None, None)


def test_done_callback():
    import ubelt as ub
    self1 = ub.Executor(mode='serial', max_workers=0)
    with self1:
        jobs = []
        for i in range(10):
            jobs.append(self1.submit(sum, [i + 1, i]))
        for job in jobs:
            job.add_done_callback(lambda x: print('done callback got x = {}'.format(x)))
            result = job.result()
            print('result = {!r}'.format(result))


def _killable_worker(kill_fpath):
    """
    An infinite loop that we can kill by writing a sentinel value to disk
    """
    import ubelt as ub
    timer = ub.Timer().tic()
    while True:
        # Don't want for too long
        if timer.toc() > 10:
            return
        if kill_fpath.exists():
            return


def _sleepy_worker(seconds, loops=100):
    """
    An infinite loop that we can kill by writing a sentinel value to disk
    """
    import time
    start_time = time.monotonic()
    while True:
        time.sleep(seconds / loops)
        elapsed = time.monotonic() - start_time
        if elapsed > seconds:
            return elapsed


def test_as_completed_timeout():
    """
    xdoctest ~/code/ubelt/tests/test_futures.py test_as_completed_timeout
    """
    from concurrent.futures import TimeoutError

    import ubelt as ub
    import uuid
    kill_fname = str(uuid.uuid4()) + '.signal'

    # modes = ['thread', 'process', 'serial']
    modes = ['thread', 'process']

    timeout = 0.1
    dpath = ub.Path.appdir('ubelt', 'tests', 'futures', 'timeout').ensuredir()
    kill_fpath = dpath / kill_fname

    for mode in modes:
        jobs = ub.JobPool(mode=mode, max_workers=2)
        with jobs:
            print('Submitting')
            timer = ub.Timer().tic()

            jobs.submit(_sleepy_worker, seconds=1e-1)
            print('Submit job: ' + str(timer.toc()))
            jobs.submit(_sleepy_worker, seconds=2e-1)
            print('Submit job: ' + str(timer.toc()))
            jobs.submit(_sleepy_worker, seconds=3e-1)
            print('Submit job: ' + str(timer.toc()))
            jobs.submit(_killable_worker, kill_fpath)
            print('Submit job: ' + str(timer.toc()))
            jobs.submit(_killable_worker, kill_fpath)
            print('Submit job: ' + str(timer.toc()))
            jobs.submit(_killable_worker, kill_fpath)
            print('Submit job: ' + str(timer.toc()))
            jobs.submit(_killable_worker, kill_fpath)
            print('Submit job: ' + str(timer.toc()))

            print('Finished submit')
            timer.tic()
            try:
                completed_iter = jobs.as_completed(timeout=timeout)
                timer.tic()
                for job in completed_iter:
                    print('Collect job: ' + str(timer.toc()))
                    try:
                        job.result()
                    except Exception as ex:
                        print(f'ex={ex}')
                        ...
                    # print('job = {}'.format(ub.urepr(job, nl=1)))
                    timer.tic()
                    ...
            except TimeoutError as ex:
                print(f'We got a timeout ex={ex}')

            print('Handled timeout: ' + str(timer.toc()))
            print('We cant escape this context until the jobs finish')
            print([j._state for j in jobs.jobs])
            kill_fpath.touch()
            print([j._state for j in jobs.jobs])
        print([j._state for j in jobs.jobs])

        print('Cleanup')
        kill_fpath.delete()
        print('End of function')


if __name__ == '__main__':
    """
    CommandLine:
        python ~/code/ubelt/tests/test_futures.py
    """
    test_as_completed_timeout()
    # import xdoctest
    # xdoctest.doctest_module(__file__)



from typing import Dict
from typing import List
from ubelt.util_const import NoParam
from ubelt.util_const import NoParamType
from typing import Any
from typing import Callable
from typing import Tuple
from os import PathLike
from typing import Any, TypeVar

Hasher = TypeVar("Hasher")
HASH_VERSION: int
DEFAULT_ALPHABET: List[str]


def b(s: str) -> bytes:
    ...


DEFAULT_HASHER: Callable


class _Hashers:
    algos: Dict[str, object]
    aliases: Dict[str, str]

    def __init__(self) -> None:
        ...

    def available(self) -> List[str]:
        ...

    def __contains__(self, key: str) -> bool:
        ...

    def lookup(self, hasher: NoParamType | str | Any) -> Callable:
        ...


class HashableExtensions:
    iterable_checks: List[Callable]

    def __init__(self) -> None:
        ...

    def register(self, hash_types: type | Tuple[type]) -> Callable:
        ...

    def lookup(self, data: object) -> Callable:
        ...

    def add_iterable_check(self, func: Callable) -> Callable:
        ...


class _HashTracer:
    sequence: List[bytes]

    def __init__(self) -> None:
        ...

    def update(self, item: bytes) -> None:
        ...

    def hexdigest(self) -> bytes:
        ...


def hash_data(data: object,
              hasher: str | Hasher | NoParamType = NoParam,
              base: List[str] | str | NoParamType = NoParam,
              types: bool = False,
              convert: bool = False,
              extensions: HashableExtensions | None = None) -> str:
    ...


def hash_file(fpath: PathLike,
              blocksize: int = 1048576,
              stride: int = 1,
              maxbytes: int | None = None,
              hasher: str | Hasher | NoParamType = NoParam,
              base: List[str] | int | str | NoParamType = NoParam) -> str:
    ...



import random
import ubelt as ub
import itertools as it


def test_operators():
    """
    CommandLine:
        python ~/code/ubelt/ubelt/tests/test_oset.py test_operators
    """
    rng = random.Random(0)

    def random_oset(rng, a=20, b=20):
        return ub.OrderedSet(rng.randint(0, a) for _ in range(b))

    def check_results(*results, **kw):
        name = kw.get('name', 'set test')
        datas = kw.get('datas', [])
        if not ub.allsame(results):
            raise AssertionError('Not all same {} for {} with datas={}'.format(
                results, name, datas))
        for a, b in it.combinations(results, 2):
            if not isinstance(a, (bool, int)):
                assert a is not b, name + ' should all be different items'

    def operator_tests(data1, data2):
        result1 = data1.copy()
        print('====')
        print('data1 = {!r}'.format(data1))
        print('data2 = {!r}'.format(data2))
        result1.intersection_update(data2)
        result2 = (data1 & data2)
        result3 = (data1.intersection(data2))
        print('result1 = {!r} result1.intersection_update(data2)'.format(result1))
        print('result2 = {!r} (data1 & data2) '.format(result2))
        print('result3 = {!r} (data1.intersection(data2))'.format(result3))
        check_results(result1, result2, result3, datas=(data1, data2),
                      name='isect')

        result1 = data1.copy()
        result1.difference_update(data2)
        result2 = (data1 - data2)
        result3 = (data1.difference(data2))
        check_results(result1, result2, result3, datas=(data1, data2),
                      name='-')

        result1 = data1.copy()
        result1.symmetric_difference_update(data2)
        result2 = (data1 ^ data2)
        result3 = (data1.symmetric_difference(data2))
        check_results(result1, result2, result3, datas=(data1, data2),
                      name='xor')

        result1 = data1.copy()
        result1.update(data2)
        result2 = (data1 | data2)
        result3 = (data1.union(data2))
        check_results(result1, result2, result3, datas=(data1, data2),
                      name='union')

        result1 = data1 <= data2
        result2 = data1.issubset(data2)
        result3 = set(data1).issubset(set(data2))
        check_results(result1, result2, result3, datas=(data1, data2),
                      name='subset')

        result1 = data1 >= data2
        result2 = data1.issuperset(data2)
        result3 = set(data1).issuperset(set(data2))
        check_results(result1, result2, result3, datas=(data1, data2),
                      name='superset')

        result1 = data1.isdisjoint(data2)
        result2 = len(data1.intersection(data2)) == 0
        check_results(result1, result2, datas=(data1, data2),
                      name='disjoint')

    # run tests on standard test cases
    data1 = ub.OrderedSet([5, 3, 1, 4])
    data2 = ub.OrderedSet([1, 4])
    operator_tests(data1, data2)
    operator_tests(data2, data1)

    data1 = ub.OrderedSet([])
    data2 = ub.OrderedSet([])
    operator_tests(data1, data2)

    data1 = ub.OrderedSet([3, 1, 2])
    data2 = ub.OrderedSet([])
    operator_tests(data1, data2)
    operator_tests(data2, data1)

    # run tests on random test cases
    for _ in range(10):
        data1 = random_oset(rng)
        data2 = random_oset(rng)

        operator_tests(data1, data2)
        operator_tests(data2, data1)


def test_equality():

    def check(a, b):
        # Self checks
        assert a == a
        assert a >= a
        assert a <= a
        assert not a < a
        assert not a > a
        assert not a != a

        # Lesser checks
        assert a < b
        assert a <= b
        assert a != b
        assert not a == b

        # Greater checks
        assert b > a
        assert b >= a
        assert b != a
        assert not b == a

    a = ub.oset([])
    b = ub.oset([1])
    c = ub.oset([1, 2])
    d = ub.oset([1, 2, 3])

    check(a, b)
    check(b, c)
    check(c, d)
    check(a, d)
    check(a, d)


if __name__ == '__main__':
    r"""
    CommandLine:
        python ~/code/ubelt/ubelt/tests/test_oset.py test_equality
        pytest ~/code/ubelt/ubelt/tests/test_oset.py
    """
    import xdoctest
    xdoctest.doctest_module(__file__)
